# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a channel is deleted"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects import Channel

if TYPE_CHECKING:
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def channel_delete_middleware(
    self: Client, gateway: Gateway, payload: GatewayDispatch
):
    """|coro|

    Middleware for the ``on_channel_delete`` event.

    Parameters
    ----------
    payload : :class:`pincer.core.gateway.GatewayDispatch`
        The data received from the channel delete event.
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.objects.guild.channel.Channel`]
        ``on_channel_delete`` and a ``Channel``
    """

    channel = Channel.from_dict(payload.data)

    guild = self.guilds.get(channel.guild_id)
    if guild:
        guild.channels = [c for c in guild.channels if c.id != channel.id]

    self.channels.pop(channel.id, None)

    return "on_channel_delete", channel


def export():
    return channel_delete_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import IntFlag
from typing import Tuple, Optional


class PermissionEnum(IntFlag):
    """
    Represents the permissions for a guild.
    """

    CREATE_INSTANT_INVITE = 1 << 0
    KICK_MEMBERS = 1 << 1
    BAN_MEMBERS = 1 << 2
    ADMINISTRATOR = 1 << 3
    MANAGE_CHANNELS = 1 << 4
    MANAGE_GUIlD = 1 << 5
    ADD_REACTIONS = 1 << 6
    VIEW_AUDIT_LOG = 1 << 7
    PRIORITY_SPEAKER = 1 << 8
    STREAM = 1 << 9
    VIEW_CHANNEL = 1 << 10
    SEND_MESSAGES = 1 << 11
    SEND_TTS_MESSAGES = 1 << 12
    MANAGE_MESSAGES = 1 << 13
    EMBED_LINKS = 1 << 14
    ATTACH_FILES = 1 << 15
    READ_MESSAGE_HISTORY = 1 << 16
    MENTION_EVERYONE = 1 << 17
    USE_EXTERNAL_EMOJIS = 1 << 18
    VIEW_GUILD_INSIGHTS = 1 << 19
    CONNECT = 1 << 20
    SPEAK = 1 << 21
    MUTE_MEMBERS = 1 << 22
    DEAFEN_MEMBERS = 1 << 23
    MOVE_MEMBERS = 1 << 24
    USE_VAD = 1 << 25
    CHANGE_NICKNAME = 1 << 26
    MANAGE_NICKNAMES = 1 << 27
    MANAGE_ROLES = 1 << 28
    MANAGE_WEBHOOKS = 1 << 29
    MANAGE_EMOJIS_AND_STICKERS = 1 << 30
    USE_APPLICATION_COMMANDS = 1 << 31
    REQUEST_TO_SPEAK = 1 << 32
    MANAGE_EVENTS = 1 << 33
    MANAGE_THREADS = 1 << 34
    CREATE_PUBLIC_THREADS = 1 << 35
    CREATE_PRIVATE_THREADS = 1 << 36
    USE_EXTERNAL_STICKERS = 1 << 37
    SEND_MESSAGES_IN_THREADS = 1 << 38
    START_EMBEDDED_ACTIVITIES = 1 << 39
    MODERATE_MEMBERS = 1 << 40


@dataclass
class Permissions:
    """
    Allows for easier access to the permissions

    Parameters
    ----------
    create_instant_invite: :class:Optional[:class:`bool`]
        Allows creation of instant invites
    kick_members: :class:Optional[:class:`bool`]
        Allows kicking members
    ban_members: :class:Optional[:class:`bool`]
        Allows banning members
    administrator: :class:Optional[:class:`bool`]
        Allows all permissions and bypasses channel permission overwrites
    manage_channels: :class:Optional[:class:`bool`]
        Allows management and editing of channels
    manage_guild: :class:Optional[:class:`bool`]
        Allows management and editing of the guild
    add_reactions: :class:Optional[:class:`bool`]
        Allows for the addition of reactions to messages
    view_audit_log: :class:Optional[:class:`bool`]
        Allows for viewing of audit logs
    priority_speaker: :class:Optional[:class:`bool`]
        Allows for using priority speaker in a voice channel
    stream: :class:Optional[:class:`bool`]
        Allows the user to go live
    view_channel: :class:Optional[:class:`bool`]
        Allows guild members to view a channel, which includes reading messages in text channels
    send_messages: :class:Optional[:class:`bool`]
        Allows for sending messages in a channel (does not allow sending messages in threads)
    send_tts_messages: :class:Optional[:class:`bool`]
        Allows for sending of tts messages
    manage_messages: :class:Optional[:class:`bool`]
        Allows for deletion of other users messages
    embed_links: :class:Optional[:class:`bool`]
        Links sent by users with this permission will be auto-embedded
    attach_files: :class:Optional[:class:`bool`]
        Allows for uploading images and files
    read_message_history: :class:Optional[:class:`bool`]
        Allows for reading of message history
    mention_everyone: :class:Optional[:class:`bool`]
        Allows for using the @everyone tag to notify all users in a channel, and the @here tag to notify all online users in a channel
    use_external_emojis: :class:Optional[:class:`bool`]
        Allows the usage of custom emojis from other servers
    view_guild_insights: :class:Optional[:class:`bool`]
        Allows for viewing of guild insights
    connect: :class:Optional[:class:`bool`]
        Allows for joining of a voice channel
    speak: :class:Optional[:class:`bool`]
        Allows for speaking in a voice channel
    mute_members: :class:Optional[:class:`bool`]
        Allows for muting members in a voice channel
    deafen_members: :class:Optional[:class:`bool`]
        Allows for deafening of members in a voice channel
    move_members: :class:Optional[:class:`bool`]
        Allows for moving of members between voice channels
    use_vad: :class:Optional[:class:`bool`]
        Allows for using voice activity detection in a voice channel
    change_nickname: :class:Optional[:class:`bool`]
        Allows for modification of own nickname
    manage_nicknames: :class:Optional[:class:`bool`]
        Allows for modification of other users nicknames
    manage_roles: :class:Optional[:class:`bool`]
        Allows for management and editing of roles
    manage_webhooks: :class:Optional[:class:`bool`]
        Allows for management and editing of webhooks
    manage_emojis_and_stickers: :class:Optional[:class:`bool`]
        Allows for management and editing of emojis and stickers
    use_application_commands: :class:Optional[:class:`bool`]
        Allows for using application-specific commands
    request_to_speak: :class:Optional[:class:`bool`]
        Allows for requesting to speak in a voice channel
    manage_events: :class:Optional[:class:`bool`]
        Allows for management and editing of events
    manage_threads: :class:Optional[:class:`bool`]
        Allows for management and editing of threads
    create_public_threads: :class:Optional[:class:`bool`]
        Allows for the creation of public threads
    create_private_threads: :class:Optional[:class:`bool`]
        Allows for the creation of private threads
    use_external_stickers: :class:Optional[:class:`bool`]
        Allows for the usage of stickers from other servers
    send_messages_in_threads: :class:Optional[:class:`bool`]
        Allows for sending messages in threads
    start_embedded_activities: :class:Optional[:class:`bool`]
        Allows for starting of embedded activities
    moderate_members: :class:Optional[:class:`bool`]
        Allows for moderation of members in a guild
    """

    create_instant_invite: Optional[bool] = None
    kick_members: Optional[bool] = None
    ban_members: Optional[bool] = None
    administrator: Optional[bool] = None
    manage_channels: Optional[bool] = None
    manage_guild: Optional[bool] = None
    add_reactions: Optional[bool] = None
    view_audit_log: Optional[bool] = None
    priority_speaker: Optional[bool] = None
    stream: Optional[bool] = None
    view_channel: Optional[bool] = None
    send_messages: Optional[bool] = None
    send_tts_messages: Optional[bool] = None
    manage_messages: Optional[bool] = None
    embed_links: Optional[bool] = None
    attach_files: Optional[bool] = None
    read_message_history: Optional[bool] = None
    mention_everyone: Optional[bool] = None
    use_external_emojis: Optional[bool] = None
    view_guild_insights: Optional[bool] = None
    connect: Optional[bool] = None
    speak: Optional[bool] = None
    mute_members: Optional[bool] = None
    deafen_members: Optional[bool] = None
    move_members: Optional[bool] = None
    use_vad: Optional[bool] = None
    change_nickname: Optional[bool] = None
    manage_nicknames: Optional[bool] = None
    manage_roles: Optional[bool] = None
    manage_webhooks: Optional[bool] = None
    manage_emojis_and_stickers: Optional[bool] = None
    use_application_commands: Optional[bool] = None
    request_to_speak: Optional[bool] = None
    manage_events: Optional[bool] = None
    manage_threads: Optional[bool] = None
    create_public_threads: Optional[bool] = None
    create_private_threads: Optional[bool] = None
    use_external_stickers: Optional[bool] = None
    send_messages_in_threads: Optional[bool] = None
    start_embedded_activities: Optional[bool] = None
    moderate_members: Optional[bool] = None

    def __setattr__(self, name: str, value: Optional[bool]) -> None:
        if not isinstance(value, bool) and value is not None:
            raise ValueError(f"Permission {name!r} must be a boolean or None")
        return super().__setattr__(name, value)

    @classmethod
    def from_ints(cls, allow: int, deny: int) -> Permissions:
        """
        Create a Permission object from an integer representation of the permissions (deny and allow)

        Parameters
        ----------
        allow: :class:`int`
            The integer representation of the permissions that are allowed
        deny: :class:`int`
            The integer representation of the permissions that are denied
        """
        clsobj = cls()

        for enum in PermissionEnum:
            value = None
            if enum.value & allow:
                value = True
            elif enum.value & deny:
                value = False

            setattr(clsobj, enum.name.lower(), value)

        return clsobj

    def to_ints(self) -> Tuple[int]:
        """
        Convert the Permission object to an integer representation of the permissions (deny and allow)

        Returns
        -------
        :class:`Tuple[:class:`int`]`
            The integer representation of the permissions that are allowed and denied
        """
        allow = 0
        deny = 0
        for enum in PermissionEnum:
            if getattr(self, enum.name.lower()):
                allow |= enum.value
            elif getattr(self, enum.name.lower()) is False:
                deny |= enum.value

        return allow, deny

    @property
    def allow(self) -> int:
        """
        Returns the integer representation of the permissions that are allowed
        """
        allow = 0
        for enum in PermissionEnum:
            if getattr(self, enum.name.lower()):
                allow |= enum.value

        return allow

    @property
    def deny(self) -> int:
        """
        Returns the integer representation of the permissions that are denied
        """
        deny = 0
        for enum in PermissionEnum:
            if getattr(self, enum.name.lower()) is False:
                deny |= enum.value

        return deny



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Tuple

    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def payload_middleware(
    self: Client,
    gateway: Gateway,
    payload: GatewayDispatch,
) -> Tuple[str, GatewayDispatch]:
    """Invoked when anything is received from gateway.


    Parameters
    ----------
    payload : :class:`pincer.core.gateway.GatewayDispatch`
        The data received from the ready event.
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.core.gateway.GatewayDispatch`]
        ``on_payload`` and a ``payload``
    """
    return "on_payload", payload


def export():
    return payload_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .application import Application
from .command import (
    AppCommandInteractionDataOption,
    AppCommandOptionChoice,
    AppCommandOption,
    AppCommand,
    InteractableStructure,
)
from .command_types import AppCommandType, AppCommandOptionType
from .intents import Intents
from .interaction_base import CallbackType, InteractionType, MessageInteraction
from .interaction_flags import InteractionFlags
from .interactions import ResolvedData, InteractionData, Interaction
from .mentionable import Mentionable
from .session_start_limit import SessionStartLimit
from .throttle_scope import ThrottleScope
from .throttling import ThrottleInterface, DefaultThrottleHandler


__all__ = (
    "AppCommand",
    "AppCommandInteractionDataOption",
    "AppCommandOption",
    "AppCommandOptionChoice",
    "AppCommandOptionType",
    "AppCommandType",
    "Application",
    "CallbackType",
    "DefaultThrottleHandler",
    "Intents",
    "InteractableStructure",
    "Interaction",
    "InteractionData",
    "InteractionFlags",
    "InteractionType",
    "Mentionable",
    "MessageInteraction",
    "ResolvedData",
    "SessionStartLimit",
    "ThrottleInterface",
    "ThrottleScope",
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

import os
from base64 import b64encode
from dataclasses import dataclass
from io import BytesIO
from json import dumps
from typing import TYPE_CHECKING

from aiohttp import FormData, Payload

from ...exceptions import ImageEncodingError

if TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Tuple

    IMAGE_TYPE = Any

PILLOW_IMPORT = True


try:
    from PIL.Image import Image

    if TYPE_CHECKING:
        IMAGE_TYPE = Image
except (ModuleNotFoundError, ImportError):
    PILLOW_IMPORT = False


def create_form(
    json_payload: Dict[Any], files: List[File]
) -> Tuple[str, Payload]:
    """
    Creates an aiohttp payload from an array of File objects.

    json_payload : Dict[Any]
        The json part of the request
    files : List[`~pincer.objects.message.file.File`]
        A list of files to be used in the request.

    Returns
    -------
    Tuple[str, :class:`aiohttp.Payload`]
        The content type and the payload to be sent in an HTTP request.
    """
    form = FormData()
    form.add_field("payload_json", dumps(json_payload))

    for file in files:
        if not file.filename:
            raise ImageEncodingError(
                "A filename is required for uploading attachments"
            )

        form.add_field("file", file.content, filename=file.filename)

    payload = form()
    return payload.headers["Content-Type"], payload


def _get_file_extension(filename: str) -> Optional[str]:
    """
    Returns the file extension from a str if it exists, otherwise
    return :data:`None`.

    filename : str
        The filename

    Returns
    -------
    Optional[:class:`str`]
        The file extension or :data:`None`
    """
    path = os.path.splitext(filename)
    if len(path) >= 2:
        return path[1][1:]
    return None


@dataclass
class File:
    """A file that is prepared by the user to be sent to the discord
    API.

    Attributes
    ----------
    content: :class:`bytes`
        File bytes.
    filename: :class:`str`
        The name of the file when it's uploaded to discord.
    """

    content: bytes
    image_format: str
    filename: Optional[str] = None

    @classmethod
    def from_file(cls, filepath: str, filename: str = None) -> File:
        """Make a ``File`` object from a file stored locally.

        Parameters
        ----------
        filepath: :class:`str`
            The path to the file you want to send. Must be string. The file's
            name in the file path is used as the name when uploaded to discord
            by default.

        filename: :class:`str`
            The name of the file. Will override the default name.
            |default| ``os.path.basename(filepath)``

        Returns
        -------
        :class:`~pincer.objects.message.file.File`
            The new file object.
        """
        with open(filepath, "rb") as data:
            file = data.read()

        return cls(
            content=file,
            image_format=_get_file_extension(filename),
            filename=filename or os.path.basename(filepath),
        )

    @classmethod
    def from_pillow_image(
        cls,
        img: IMAGE_TYPE,
        filename: Optional[str] = None,
        image_format: Optional[str] = None,
        **kwargs,
    ) -> File:
        """Creates a file object from a PIL image
        Supports GIF, PNG, JPEG, and WEBP.

        Parameters
        ----------
        img: :class:`~pil:PIL.Image.Image`
            Pillow image object.
        filename:
            The filename to be used when uploaded to discord. The extension is
            used as image_format unless otherwise specified.
        image_format:
            The image_format to be used if you want to override the file
            extension.

        Returns
        -------
        :class:`~pincer.objects.message.file.File`
            The new file object.

        Raises
        ------
        ModuleNotFoundError:
            ``Pillow`` is not installed
        """
        if not PILLOW_IMPORT:
            raise ModuleNotFoundError(
                "The `Pillow` library is required for sending and converting "
                "pillow images,"
            )

        if image_format is None:
            image_format = _get_file_extension(filename)

        if image_format == "jpg":
            image_format = "jpeg"

        # https://stackoverflow.com/questions/33101935/convert-pil-image-to-byte-array
        # Credit goes to second answer
        img_byte_arr = BytesIO()
        img.save(img_byte_arr, format=image_format)
        img_bytes = img_byte_arr.getvalue()

        return cls(
            content=img_bytes, image_format=image_format, filename=filename
        )

    @property
    def uri(self) -> str:
        """
        Returns
        -------
        str
            The uri for the image.
            See `<https://discord.com/developers/docs/reference#api-versioning>`_.
        """  # noqa: E501
        if self.image_format not in {"jpeg", "png", "gif"}:
            raise ImageEncodingError(
                'Only image types "jpeg", "png", and "gif" can be sent in'
                " an Image URI"
            )

        encoded_bytes = b64encode(self.content).decode("ascii")

        return f"data:image/{self.image_format};base64,{encoded_bytes}"

    @property
    def content_type(self):
        return f"image/{self.image_format}"



"""
Pincer library
====================
An asynchronous python API wrapper meant to replace discord.py

Copyright Pincer 2021
Full MIT License can be found in `LICENSE` at the project root.
"""

from typing import NamedTuple, Literal, Optional

from ._config import GatewayConfig
from .client import event_middleware, Client, Bot
from .cog import Cog
from .commands import command, ChatCommandHandler
from .exceptions import (
    PincerError,
    InvalidPayload,
    UnhandledException,
    NoExportMethod,
    CogError,
    CogNotFound,
    CogAlreadyExists,
    NoValidSetupMethod,
    TooManySetupArguments,
    NoCogManagerReturnFound,
    CommandError,
    CommandCooldownError,
    CommandIsNotCoroutine,
    CommandAlreadyRegistered,
    CommandDescriptionTooLong,
    TooManyArguments,
    InvalidArgumentAnnotation,
    CommandReturnIsEmpty,
    InvalidCommandGuild,
    InvalidCommandName,
    InvalidEventName,
    InvalidUrlError,
    EmbedFieldError,
    TaskError,
    TaskAlreadyRunning,
    TaskCancelError,
    TaskIsNotCoroutine,
    TaskInvalidDelay,
    DispatchError,
    DisallowedIntentsError,
    InvalidTokenError,
    HeartbeatError,
    UnavailableGuildError,
    HTTPError,
    NotModifiedError,
    BadRequestError,
    UnauthorizedError,
    ForbiddenError,
    NotFoundError,
    MethodNotAllowedError,
    RateLimitError,
    GatewayError,
    ServerError,
    EmbedOverflow,
    ImageEncodingError,
)
from .objects import Intents

__package__ = "pincer"
__title__ = "Pincer library"
__description__ = "Discord API wrapper rebuild from scratch."
__author__ = "Sigmanificient, Arthurdw"
__email__ = "contact@pincer.dev"
__license__ = "MIT"

ReleaseType = Optional[Literal["alpha", "beta", "candidate", "final", "dev"]]


class VersionInfo(NamedTuple):
    """A Class representing the version of the Pincer library."""

    major: int
    minor: int
    micro: int

    release_level: ReleaseType = None
    serial: int = 0

    def __repr__(self) -> str:
        return f"{self.major}.{self.minor}.{self.micro}" + (
            f"-{self.release_level}{self.serial}"
            * (self.release_level is not None)
        )


version_info = VersionInfo(0, 16, 1)
__version__ = repr(version_info)

__all__ = (
    "BadRequestError",
    "Bot",
    "ChatCommandHandler",
    "Client",
    "Cog",
    "CogAlreadyExists",
    "CogError",
    "CogNotFound",
    "CommandAlreadyRegistered",
    "CommandCooldownError",
    "CommandDescriptionTooLong",
    "CommandError",
    "CommandIsNotCoroutine",
    "CommandReturnIsEmpty",
    "DisallowedIntentsError",
    "DispatchError",
    "EmbedFieldError",
    "EmbedOverflow",
    "ForbiddenError",
    "GatewayConfig",
    "GatewayError",
    "HTTPError",
    "HeartbeatError",
    "ImageEncodingError",
    "Intents",
    "InvalidArgumentAnnotation",
    "InvalidCommandGuild",
    "InvalidCommandName",
    "InvalidEventName",
    "InvalidPayload",
    "InvalidTokenError",
    "InvalidUrlError",
    "MethodNotAllowedError",
    "NoCogManagerReturnFound",
    "NoExportMethod",
    "NoValidSetupMethod",
    "NotFoundError",
    "NotModifiedError",
    "PincerError",
    "RateLimitError",
    "ServerError",
    "TaskAlreadyRunning",
    "TaskCancelError",
    "TaskError",
    "TaskInvalidDelay",
    "TaskIsNotCoroutine",
    "TooManyArguments",
    "TooManySetupArguments",
    "UnauthorizedError",
    "UnavailableGuildError",
    "UnhandledException",
    "__author__",
    "__email__",
    "__package__",
    "__title__",
    "__version__",
    "command",
    "event_middleware",
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.
from __future__ import annotations

from sys import modules
from typing import TypeVar, Callable, Coroutine, Any, Union, Optional


class MissingType:
    """Type class for missing attributes and parameters."""

    def __repr__(self):
        return "<MISSING>"

    def __bool__(self) -> bool:
        return False

    def __eq__(self, __o: object) -> bool:
        return __o is MISSING

    def __hash__(self) -> int:
        # By returning the hash of None, when searching for a command it doesn't know
        # what kind of "nothing" it is.
        return hash(None)


MISSING = MissingType()


T = TypeVar("T")

JSONSerializable = TypeVar(
    "JSONSerializable", str, int, float, list, dict, bool, None
)


# Represents a value which is optionally returned from the API
APINullable = Union[T, MissingType]


#: Represents a coroutine.
Coro = TypeVar("Coro", bound=Callable[..., Coroutine[Any, Any, Any]])

choice_value_types = Union[str, int, float]

CheckFunction = Optional[Callable[[Any], bool]]


class Singleton(type):
    # Thanks to this stackoverflow answer (method 3):
    # https://stackoverflow.com/q/6760685/12668716
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(
                *args, **kwargs
            )
        return cls._instances[cls]


class TypeCache(metaclass=Singleton):
    cache = {}

    def __init__(self):
        # Register all known types to the cache. This gets used later
        # to auto-convert the properties to their desired type.
        lcp = modules.copy()
        for module in lcp:
            if not module.startswith("pincer"):
                continue

            TypeCache.cache.update(lcp[module].__dict__)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .attachment import Attachment
from .component import MessageComponent, ComponentType
from .context import MessageContext
from .embed import (
    Embed,
    EmbedField,
    EmbedImage,
    EmbedAuthor,
    EmbedProvider,
    EmbedThumbnail,
    EmbedVideo,
    EmbedFooter,
)
from .emoji import Emoji
from .file import File
from .message import Message
from .reaction import Reaction
from .reference import MessageReference
from .sticker import (
    StickerType,
    StickerFormatType,
    Sticker,
    StickerItem,
    StickerPack,
)
from .user_message import (
    MessageActivityType,
    MessageFlags,
    MessageType,
    MessageActivity,
    AllowedMentionTypes,
    UserMessage,
    AllowedMentions,
)

__all__ = (
    "AllowedMentionTypes",
    "AllowedMentions",
    "Attachment",
    "ComponentType",
    "Embed",
    "EmbedAuthor",
    "EmbedField",
    "EmbedFooter",
    "EmbedImage",
    "EmbedProvider",
    "EmbedThumbnail",
    "EmbedVideo",
    "Emoji",
    "File",
    "Message",
    "MessageActivity",
    "MessageActivityType",
    "MessageComponent",
    "MessageContext",
    "MessageFlags",
    "MessageReference",
    "MessageType",
    "Reaction",
    "Sticker",
    "StickerFormatType",
    "StickerItem",
    "StickerPack",
    "StickerType",
    "UserMessage",
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from typing import Generic, TypeVar, TYPE_CHECKING

from pincer.utils import APIObject

if TYPE_CHECKING:
    from typing import Any, Coroutine, List, Type, Generator, AsyncIterator

T = TypeVar("T")


class APIDataGen(Generic[T]):
    def __init__(
        self, factory: Type[T], request_func: Coroutine[Any, None, Any]
    ):

        self.fac = (
            factory if isinstance(factory, APIObject) else factory.from_dict
        )
        self.request_func = request_func

    async def __async(self) -> List[T]:
        data = await self.request_func
        return [self.fac(i) for i in data]

    def __await__(self) -> Generator[Any, None, Any]:
        return self.__async().__await__()

    async def __aiter__(self) -> AsyncIterator[List[T]]:
        for item in await self:
            yield self.fac(item)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

import logging
import signal
from asyncio import (
    iscoroutinefunction,
    ensure_future,
    create_task,
    get_event_loop,
)
from collections import defaultdict
from functools import partial
from inspect import isasyncgenfunction
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Iterable,
    OrderedDict,
    Tuple,
    Union,
    overload,
    TYPE_CHECKING,
)


from .cog import CogManager
from .commands.interactable import Interactable
from .objects.app.command import InteractableStructure

from . import __package__
from .commands import ChatCommandHandler
from .core import HTTPClient
from .core.gateway import GatewayInfo, Gateway

from .exceptions import InvalidEventName, GatewayConnectionError

from .middleware import middleware
from .objects import (
    Role,
    Channel,
    DefaultThrottleHandler,
    User,
    Guild,
    Intents,
    GuildTemplate,
    StickerPack,
    UserMessage,
    Connection,
    File,
    StageInstance,
    PrivacyLevel,
)
from .objects.guild.channel import GroupDMChannel
from .utils import APIObject
from .utils.conversion import remove_none
from .utils.event_mgr import EventMgr
from .utils.extraction import get_index
from .utils.insertion import should_pass_cls, should_pass_gateway
from .utils.shards import calculate_shard_id
from .utils.types import CheckFunction
from .utils.types import Coro

if TYPE_CHECKING:
    from .utils.snowflake import Snowflake
    from .core.dispatch import GatewayDispatch
    from .objects.app.throttling import ThrottleInterface
    from .objects.guild import Webhook

    from collections.abc import AsyncIterator

_log = logging.getLogger(__package__)

MiddlewareType = Optional[Union[Coro, Tuple[str, List[Any], Dict[str, Any]]]]

_event = Union[str, InteractableStructure[None]]
_events: Dict[str, Optional[Union[List[_event], _event]]] = defaultdict(list)


def event_middleware(call: str, *, override: bool = False):
    """Middleware are methods which can be registered with this decorator.
    These methods are invoked before any ``on_`` event.
    As the ``on_`` event is the final call.

    A default call exists for all events, but some might already be in
    use by the library.

    If you know what you are doing, you can override these default
    middleware methods by passing the override parameter.

    The method to which this decorator is registered must be a coroutine,
    and it must return a tuple with the following format:

    .. code-block:: python

        tuple(
            key for next middleware or final event [str],
            args for next middleware/event which will be passed as *args
                [list(Any)],
            kwargs for next middleware/event which will be passed as
                **kwargs [dict(Any)]
        )

    Two parameters are passed to the middleware. The first parameter is
    the current socket connection with and the second one is the payload
    parameter which is of type :class:`~.core.dispatch.GatewayDispatch`.
    This contains the response from the discord API.

    :Implementation example:

    .. code-block:: python3

         @event_middleware("ready", override=True)
         async def custom_ready(_, payload: GatewayDispatch):
             return "on_ready", [
                 User.from_dict(payload.data.get("user"))
             ]

         @Client.event
         async def on_ready(bot: User):
             print(f"Signed in as {bot}")

    Parameters
    ----------
    call : :class:`str`
        The call that the function should tie to
    override : :class:`bool`
        If it should override default middleware,
        usually shouldn't be used |default| :data:`False`
    """

    def decorator(func: Coro):
        if override:
            _log.warning(
                "Middleware overriding has been enabled for `%s`."
                " This might cause unexpected behavior.",
                call,
            )

        if not override and callable(_events.get(call)):
            raise RuntimeError(
                f"Middleware event with call `{call}` has "
                "already been registered"
            )

        async def wrapper(cls, gateway: Gateway, payload: GatewayDispatch):
            _log.debug("`%s` middleware has been invoked", call)

            return await func(cls, gateway, payload)

        _events[call] = wrapper
        return wrapper

    return decorator


for event, middleware_ in middleware.items():
    event_middleware(event)(middleware_)


class Client(Interactable, CogManager):
    """The client is the main instance which is between the programmer
    and the discord API.

    This client represents your bot.

    Attributes
    ----------
    bot: :class:`~objects.user.user.User`
        The user object of the client
    received_message: :class:`str`
        The default message which will be sent when no response is given.
    http: :class:`~core.http.HTTPClient`
        The http client used to communicate with the discord API

    Parameters
    ----------
    token : :class:`str`
        The token to login with your bot from the developer portal
    received : Optional[:class:`str`]
        The default message which will be sent when no response is given.
        |default| :data:`None`
    intents : Optional[Union[Iterable[:class:`~objects.app.intents.Intents`], :class:`~objects.app.intents.Intents`]]
        The discord intents to use |default| :data:`Intents.all()`
    throttler : Optional[:class:`~objects.app.throttling.ThrottleInterface`]
        The cooldown handler for your client,
        defaults to :class:`~.objects.app.throttling.DefaultThrottleHandler`
        *(which uses the WindowSliding technique)*.
        Custom throttlers must derive from
        :class:`~pincer.objects.app.throttling.ThrottleInterface`.
        |default| :class:`~pincer.objects.app.throttling.DefaultThrottleHandler`
    """  # noqa: E501

    def __init__(
        self,
        token: str,
        *,
        received: str = None,
        intents: Intents = None,
        throttler: ThrottleInterface = DefaultThrottleHandler,
        reconnect: bool = True,
    ):
        def sigint_handler(_signal, _frame):
            _log.info("SIGINT received, shutting down...")

            # A print statement to make sure the user sees the message
            print("Closing the client loop, this can take a few seconds...")

            create_task(self.http.close())
            if self.loop.is_running():
                self.loop.stop()

        signal.signal(signal.SIGINT, sigint_handler)

        if isinstance(intents, Iterable):
            intents = sum(intents)

        if intents is None:
            intents = Intents.all()

        self.intents = intents
        self.reconnect = reconnect
        self.token = token

        self.bot: Optional[User] = None
        self.received_message = received or "Command arrived successfully!"
        self.http = HTTPClient(token)
        APIObject.bind_client(self)

        self.throttler = throttler

        async def get_gateway():
            return GatewayInfo.from_dict(await self.http.get("gateway/bot"))

        self.loop = get_event_loop()
        self.event_mgr = EventMgr(self.loop)

        self.gateway: GatewayInfo = self.loop.run_until_complete(get_gateway())
        self.shards: OrderedDict[int, Gateway] = OrderedDict()

        # The guild and channel value is only registered if the Client has the GUILDS
        # intent.
        self.guilds: Dict[Snowflake, Optional[Guild]] = {}
        self.channels: Dict[Snowflake, Optional[Channel]] = {}

        ChatCommandHandler.managers.append(self)

        super().__init__()

    @property
    def chat_commands(self) -> List[str]:
        """List[:class:`str`]: List of chat commands

        Get a list of chat command calls which have been registered in
        the :class:`~pincer.commands.ChatCommandHandler`\\.
        """
        return [
            cmd.metadata.name for cmd in ChatCommandHandler.register.values()
        ]

    @property
    def guild_ids(self) -> List[Snowflake]:
        """
        Returns a list of Guilds that the client is a part of

        Returns
        -------
        List[:class:`pincer.utils.snowflake.Snowflake`]
        """
        return self.guilds.keys()

    @staticmethod
    def event(coroutine: Coro):
        """A decorator to register a Discord gateway event listener.
        This event will get called when the client receives a new event
        update from Discord which matches the event name.

        The event name gets pulled from your method name, and this must
        start with ``on_``.
        This forces you to write clean and consistent code.

        This decorator can be used in and out of a class, and all
        event methods must be coroutines. *(async)*

        :Example usage:

        .. code-block:: python3

             # Function based
             from pincer import Client

             client = Client("token")

             @client.event
             async def on_ready():
                 print(f"Signed in as {client.bot}")

             if __name__ == "__main__":
                 client.run()

        .. code-block :: python3

             # Class based
             from pincer import Client

             class MyClient(Client):
                 @Client.event
                 async def on_ready(self):
                     print(f"Signed in as {self.bot}")

             if __name__ == "__main__":
                 MyClient("token").run()

        Raises
        ------
        TypeError
            If the function is not a coro
        InvalidEventName
            If the function name is not a valid event (on_x)
        """
        if not iscoroutinefunction(coroutine) and not isasyncgenfunction(
            coroutine
        ):
            raise TypeError(
                "Any event which is registered must be a coroutine function"
            )

        name: str = coroutine.__name__.lower()

        if not name.startswith("on_"):
            raise InvalidEventName(
                f"The event named `{name}` must start with `on_`"
            )

        if name == "on_command_error" and _events.get(name):
            raise InvalidEventName(
                f"The `{name}` event can only exist once. This is because "
                "it gets treated as a command and can have a response."
            )

        event = InteractableStructure(call=coroutine)

        _events[name].append(event)
        return event

    @staticmethod
    def get_event_coro(
        name: str,
    ) -> List[Optional[InteractableStructure[None]]]:
        """get the coroutine for an event

        Parameters
        ----------
        name : :class:`str`
            name of the event

        Returns
        -------
        List[Optional[:class:`~pincer.objects.app.command.InteractableStructure`[None]]]
        """
        calls = _events.get(name.strip().lower())

        return (
            []
            if not calls
            else [
                call
                for call in calls
                if isinstance(call, InteractableStructure)
            ]
        )

    @staticmethod
    def execute_event(
        events: List[InteractableStructure], gateway: Gateway, *args, **kwargs
    ):
        """Invokes an event.

        Parameters
        ----------
        calls: List[:class:`~pincer.objects.app.command.InteractableStructure`]
            The call (method) to which the event is registered.

        \\*args:
            The arguments for the event.

        \\*\\*kwargs:
            The named arguments for the event.
        """

        for event in events:
            call_args = args
            if should_pass_cls(event.call):
                call_args = (
                    event.manager,
                    *remove_none(args),
                )

            if should_pass_gateway(event.call):
                call_args = (call_args[0], gateway, *call_args[1:])

            ensure_future(event.call(*call_args, **kwargs))

    def run(self):
        """Start the bot."""
        ensure_future(self.start_shard(0, 1), loop=self.loop)
        self.loop.run_forever()

    def run_autosharded(self):
        """
        Runs the bot with the amount of shards specified by the Discord gateway.
        """
        num_shards = self.gateway.shards
        return self.run_shards(range(num_shards), num_shards)

    def run_shards(self, shards: Iterable, num_shards: int):
        """
        Runs shards that you specify.

        shards: Iterable
            The shards to run.
        num_shards: int
            The total amount of shards.
        """
        for shard in shards:
            ensure_future(self.start_shard(shard, num_shards), loop=self.loop)

        self.loop.run_forever()

    async def start_shard(self, shard: int, num_shards: int):
        """|coro|
        Starts a shard
        This should not be run most of the time. ``run_shards`` and ``run_autosharded``
        will likely do what you want.

        shard : int
            The number of the shard to start.
        num_shards : int
            The total number of shards.
        """

        gateway = Gateway(
            self.token,
            intents=self.intents,
            url=self.gateway.url,
            shard=shard,
            num_shards=num_shards,
        )
        await gateway.init_session()

        gateway.append_handlers(
            {
                # Gets triggered on all events
                -1: partial(self.payload_event_handler, gateway),
                # Use this event handler for opcode 0.
                0: partial(self.event_handler, gateway),
            }
        )

        self.shards[gateway.shard] = gateway
        create_task(gateway.start_loop())

    def get_shard(
        self,
        guild_id: Optional[Snowflake] = None,
        num_shards: Optional[int] = None,
    ) -> Gateway:
        """
        Returns the shard receiving events for a specified guild_id.

        ``num_shards`` is inferred from the num_shards value for the first started
        shard. If your shards do not all have the same ``num_shard`` value, you must
        specify value to get the expected result.

        Parameters
        ----------
        guild_id : Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            The guild_id of the shard to look for. If no guild id is provided, the
            shard that receives dms will be returned. |default| :data:`None`
        num_shards : Optional[:class:`int`]
            The number of shards. If no number is provided, the value will default to
            the num_shards for the first started shard. |default| :data:`None`
        """
        if not self.shards:
            raise GatewayConnectionError(
                "The client has never connected to a gateway"
            )
        if guild_id is None:
            return self.shards[0]
        if num_shards is None:
            num_shards = next(iter(self.shards.values())).num_shards
        return self.shards[calculate_shard_id(guild_id, num_shards)]

    @property
    def is_closed(self) -> bool:
        """
        Returns
        -------
        bool
            Whether the bot is closed.
        """
        return self.loop.is_running()

    def close(self):
        """
        Ensure close of the http client.
        Allow for script execution to continue.
        """
        if hasattr(self, "http"):
            create_task(self.http.close())

        self.loop.stop()

    def __del__(self):
        if self.loop.is_running():
            self.loop.stop()

        if not self.loop.is_closed():
            self.close()

    async def handle_middleware(
        self,
        payload: GatewayDispatch,
        key: str,
        gateway: Gateway,
        *args,
        **kwargs,
    ) -> Tuple[Optional[Coro], List[Any], Dict[str, Any]]:
        """|coro|

        Handles all middleware recursively. Stops when it has found an
        event name which starts with ``on_``.

        Returns a tuple where the first element is the final executor
        (so the event) its index in ``_events``.

        The second and third element are the ``*args``
        and ``**kwargs`` for the event.

        Parameters
        ----------
        payload : :class:`~pincer.core.dispatch.GatewayDispatch`
            The original payload for the event
        key : :class:`str`
            The index of the middleware in ``_events``

        Raises
        ------
        RuntimeError
            The return type must be a tuple
        RuntimeError
            Middleware has not been registered
        """
        ware: MiddlewareType = _events.get(key)
        next_call, arguments, params = ware, [], {}

        if iscoroutinefunction(ware):
            extractable = await ware(self, gateway, payload, *args, **kwargs)

            if not isinstance(extractable, tuple):
                raise RuntimeError(
                    f"Return type from `{key}` middleware must be tuple. "
                )

            next_call = get_index(extractable, 0, "")
            ret_object = get_index(extractable, 1)

        if next_call is None:
            raise RuntimeError(f"Middleware `{key}` has not been registered.")

        if next_call.startswith("on_"):
            return (next_call, ret_object)

        return await self.handle_middleware(
            payload, next_call, gateway, *arguments, **params
        )

    async def execute_error(
        self,
        error: Exception,
        gateway: Gateway,
        name: str = "on_error",
        *args,
        **kwargs,
    ):
        """|coro|

        Raises an error if no appropriate error event has been found.

        Parameters
        ----------
        error : :class:`Exception`
            The error that should be passed to the event
        name : :class:`str`
            the name of the event |default| ``on_error``

        Raises
        ------
        error
            if ``call := self.get_event_coro(name)`` is :data:`False`
        """
        if calls := self.get_event_coro(name):
            self.execute_event(calls, gateway, error, *args, **kwargs)
        else:
            raise error

    async def process_event(
        self, name: str, payload: GatewayDispatch, gateway: Gateway
    ):
        """|coro|

        Processes and invokes an event and its middleware

        Parameters
        ----------
        name : :class:`str`
            The name of the event, this is also the filename in the
            middleware directory.
        payload : :class:`~pincer.core.dispatch.GatewayDispatch`
            The payload sent from the Discord gateway, this contains the
            required data for the client to know what event it is and
            what specifically happened.
        """
        try:
            key, args = await self.handle_middleware(payload, name, gateway)
            self.event_mgr.process_events(key, args)

            if calls := self.get_event_coro(key):
                self.execute_event(calls, gateway, args)

        except Exception as e:
            await self.execute_error(e, gateway)

    async def event_handler(self, gateway: Gateway, payload: GatewayDispatch):
        """|coro|

        Handles all payload events with opcode 0.

        Parameters
        ----------
        payload : :class:`~pincer.core.dispatch.GatewayDispatch`
            The payload sent from the Discord gateway, this contains the
            required data for the client to know what event it is and
            what specifically happened.
        """
        await self.process_event(payload.event_name.lower(), payload, gateway)

    async def payload_event_handler(
        self, gateway: Gateway, payload: GatewayDispatch
    ):
        """|coro|

        Special event which activates the on_payload event.

        Parameters
        ----------
        payload : :class:`~pincer.core.dispatch.GatewayDispatch`
            The payload sent from the Discord gateway, this contains the
            required data for the client to know what event it is and
            what specifically happened.
        """
        await self.process_event("payload", payload, gateway)

    @overload
    async def create_guild(
        self,
        *,
        name: str,
        region: Optional[str] = None,
        icon: Optional[str] = None,
        verification_level: Optional[int] = None,
        default_message_notifications: Optional[int] = None,
        explicit_content_filter: Optional[int] = None,
        roles: Optional[List[Role]] = None,
        channels: Optional[List[Channel]] = None,
        afk_channel_id: Optional[Snowflake] = None,
        afk_timeout: Optional[int] = None,
        system_channel_id: Optional[Snowflake] = None,
        system_channel_flags: Optional[int] = None,
    ) -> Guild:
        """Creates a guild.

        Parameters
        ----------
        name : :class:`str`
            Name of the guild (2-100 characters)
        region : Optional[:class:`str`]
            Voice region id (deprecated) |default| :data:`None`
        icon : Optional[:class:`str`]
            base64 128x128 image for the guild icon |default| :data:`None`
        verification_level : Optional[:class:`int`]
            Verification level |default| :data:`None`
        default_message_notifications : Optional[:class:`int`]
            Default message notification level |default| :data:`None`
        explicit_content_filter : Optional[:class:`int`]
            Explicit content filter level |default| :data:`None`
        roles : Optional[List[:class:`~pincer.objects.guild.role.Role`]]
            New guild roles |default| :data:`None`
        channels : Optional[List[:class:`~pincer.objects.guild.channel.Channel`]]
            New guild's channels |default| :data:`None`
        afk_channel_id : Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            ID for AFK channel |default| :data:`None`
        afk_timeout : Optional[:class:`int`]
            AFK timeout in seconds |default| :data:`None`
        system_channel_id : Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            The ID of the channel where guild notices such as welcome
            messages and boost events are posted |default| :data:`None`
        system_channel_flags : Optional[:class:`int`]
            System channel flags |default| :data:`None`

        Returns
        -------
        :class:`~pincer.objects.guild.guild.Guild`
            The created guild
        """
        ...

    async def create_guild(self, name: str, **kwargs) -> Guild:
        g = await self.http.post("guilds", data={"name": name, **kwargs})
        return await self.get_guild(g["id"])

    async def get_guild_template(self, code: str) -> GuildTemplate:
        """|coro|
        Retrieves a guild template by its code.

        Parameters
        ----------
        code : :class:`str`
            The code of the guild template

        Returns
        -------
        :class:`~pincer.objects.guild.template.GuildTemplate`
            The guild template
        """
        return GuildTemplate.from_dict(
            await self.http.get(f"guilds/templates/{code}")
        )

    async def create_guild_from_template(
        self, template: GuildTemplate, name: str, icon: Optional[str] = None
    ) -> Guild:
        """|coro|
        Creates a guild from a template.

        Parameters
        ----------
        template : :class:`~pincer.objects.guild.template.GuildTemplate`
            The guild template
        name : :class:`str`
            Name of the guild (2-100 characters)
        icon : Optional[:class:`str`]
            base64 128x128 image for the guild icon |default| :data:`None`

        Returns
        -------
        :class:`~pincer.objects.guild.guild.Guild`
            The created guild
        """
        return Guild.from_dict(
            await self.http.post(
                f"guilds/templates/{template.code}",
                data={"name": name, "icon": icon},
            )
        )

    async def wait_for(
        self,
        event_name: str,
        check: CheckFunction = None,
        timeout: Optional[float] = None,
    ):
        """
        Parameters
        ----------
        event_name : str
            The type of event. It should start with `on_`. This is the same
            name that is used for @Client.event.
        check : CheckFunction
            This function only returns a value if this return true.
        timeout: Optional[float]
            Amount of seconds before timeout.

        Returns
        ------
        Any
            What the Discord API returns for this event.
        """
        return await self.event_mgr.wait_for(event_name, check, timeout)

    def loop_for(
        self,
        event_name: str,
        check: CheckFunction = None,
        iteration_timeout: Optional[float] = None,
        loop_timeout: Optional[float] = None,
    ):
        """
        Parameters
        ----------
        event_name : str
            The type of event. It should start with `on_`. This is the same
            name that is used for @Client.event.
        check : Callable[[Any], bool], default=None
            This function only returns a value if this return true.
        iteration_timeout: Union[float, None]
            Amount of seconds before timeout. Timeouts are for each loop.
        loop_timeout: Union[float, None]
            Amount of seconds before the entire loop times out. The generator
            will only raise a timeout error while it is waiting for an event.

        Yields
        ------
        Any
            What the Discord API returns for this event.
        """
        return self.event_mgr.loop_for(
            event_name, check, iteration_timeout, loop_timeout
        )

    async def get_guild(self, guild_id: int, with_count: bool = False) -> Guild:
        """|coro|

        Fetch a guild object by the guild identifier.

        Parameters
        ----------
        with_count: :class:bool
            Whether to include the member count in the guild object.
            Default to `False`

        guild_id : :class:`int`
            The id of the guild which should be fetched from the Discord
            gateway.

        Returns
        -------
        :class:`~pincer.objects.guild.guild.Guild`
            The guild object.
        """
        return await Guild.from_id(self, guild_id, with_count)

    async def get_user(self, _id: int) -> User:
        """|coro|

        Fetch a User from its identifier

        Parameters
        ----------
        _id : :class:`int`
            The id of the user which should be fetched from the Discord
            gateway.

        Returns
        -------
        :class:`~pincer.objects.user.user.User`
            The user object.
        """
        return await User.from_id(self, _id)

    async def get_role(self, guild_id: int, role_id: int) -> Role:
        """|coro|

        Fetch a role object by the role identifier.

        guild_id: :class:`int`
            The guild in which the role resides.

        role_id: :class:`int`
            The id of the guild which should be fetched from the Discord
            gateway.

        Returns
        -------
        :class:`~pincer.objects.guild.role.Role`
            A Role object.
        """
        return await Role.from_id(self, guild_id, role_id)

    async def get_channel(self, _id: int) -> Channel:
        """|coro|
        Fetch a Channel from its identifier. The ``get_dm_channel`` method from
        :class:`~pincer.objects.user.user.User` should be used if you need to
        create a dm_channel; using the ``send()`` method from
        :class:`~pincer.objects.user.user.User` is preferred.

        Parameters
        ----------
        _id: :class:`int`
            The id of the user which should be fetched from the Discord
            gateway.

        Returns
        -------
        :class:`~pincer.objects.guild.channel.Channel`
            A Channel object.
        """
        return await Channel.from_id(self, _id)

    async def get_message(
        self, _id: Snowflake, channel_id: Snowflake
    ) -> UserMessage:
        """|coro|
        Creates a UserMessage object

        Parameters
        ----------
        _id: :class:`~pincer.utils.snowflake.Snowflake`
            ID of the message that is wanted.
        channel_id : int
            ID of the channel the message is in.

        Returns
        -------
        :class:`~pincer.objects.message.user_message.UserMessage`
            The message object.
        """

        return await UserMessage.from_id(self, _id, channel_id)

    async def get_webhook(
        self, id: Snowflake, token: Optional[str] = None
    ) -> Webhook:
        """|coro|
        Fetch a Webhook from its identifier.

        Parameters
        ----------
        id: :class:`int`
            The id of the webhook which should be
            fetched from the Discord gateway.
        token: Optional[:class:`str`]
            The webhook token.

        Returns
        -------
        :class:`~pincer.objects.guild.webhook.Webhook`
            A Webhook object.
        """
        return await Webhook.from_id(self, id, token)

    async def get_current_user(self) -> User:
        """|coro|
        The user object of the requester's account.

        For OAuth2, this requires the ``identify`` scope,
        which will return the object *without* an email,
        and optionally the ``email`` scope,
        which returns the object *with* an email.

        Returns
        -------
        :class:`~pincer.objects.user.user.User`
        """
        return User.from_dict(await self.http.get("users/@me"))

    async def modify_current_user(
        self, username: Optional[str] = None, avatar: Optional[File] = None
    ) -> User:
        """|coro|
        Modify the requester's user account settings

        Parameters
        ----------
        username : Optional[:class:`str`]
            user's username,
            if changed may cause the user's discriminator to be randomized.
            |default| :data:`None`
        avatar : Optional[:class:`File`]
            if passed, modifies the user's avatar
            a data URI scheme of JPG, GIF or PNG
            |default| :data:`None`

        Returns
        -------
        :class:`~pincer.objects.user.user.User`
            Current modified user
        """

        avatar = avatar.uri if avatar else avatar

        user = await self.http.patch(
            "users/@me", remove_none({"username": username, "avatar": avatar})
        )
        return User.from_dict(user)

    async def get_current_user_guilds(
        self,
        before: Optional[Snowflake] = None,
        after: Optional[Snowflake] = None,
        limit: Optional[int] = None,
    ) -> AsyncIterator[Guild]:
        """|coro|
        Returns a list of partial guild objects the current user is a member of.
        Requires the ``guilds`` OAuth2 scope.

        Parameters
        ----------
        before : Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            get guilds before this guild ID
        after : Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            get guilds after this guild ID
        limit : Optional[:class:`int`]
                max number of guilds to return (1-200) |default| :data:`200`

        Yields
        ------
        :class:`~pincer.objects.guild.guild.Guild`
            A Partial Guild that the user is in
        """
        guilds = await self.http.get(
            "users/@me/guilds?"
            + (f"{before=}&" if before else "")
            + (f"{after=}&" if after else "")
            + (f"{limit=}&" if limit else "")
        )

        for guild in guilds:
            yield Guild.from_dict(guild)

    async def leave_guild(self, _id: Snowflake):
        """|coro|
        Leave a guild.

        Parameters
        ----------
        _id : :class:`~pincer.utils.snowflake.Snowflake`
            the id of the guild that the bot will leave
        """
        await self.http.delete(f"users/@me/guilds/{_id}")
        self._client.guilds.pop(_id, None)

    async def create_group_dm(
        self, access_tokens: List[str], nicks: Dict[Snowflake, str]
    ) -> GroupDMChannel:
        """|coro|
        Create a new group DM channel with multiple users.
        DMs created with this endpoint will not be shown in the Discord client

        Parameters
        ----------
        access_tokens : List[:class:`str`]
            access tokens of users that have
            granted your app the ``gdm.join`` scope

        nicks : Dict[:class:`~pincer.utils.snowflake.Snowflake`, :class:`str`]
            a dictionary of user ids to their respective nicknames

        Returns
        -------
        :class:`~pincer.objects.guild.channel.GroupDMChannel`
            group DM channel created
        """
        channel = await self.http.post(
            "users/@me/channels",
            {"access_tokens": access_tokens, "nicks": nicks},
        )

        return GroupDMChannel.from_dict(channel)

    async def get_connections(self) -> AsyncIterator[Connection]:
        """|coro|
        Returns a list of connection objects.
        Requires the ``connections`` OAuth2 scope.

        Yields
        -------
        :class:`~pincer.objects.user.connection.Connections`
            the connection objects
        """
        connections = await self.http.get("users/@me/connections")
        for conn in connections:
            yield Connection.from_dict(conn)

    async def sticker_packs(self) -> AsyncIterator[StickerPack]:
        """|coro|
        Yields sticker packs available to Nitro subscribers.

        Yields
        ------
        :class:`~pincer.objects.message.sticker.StickerPack`
            a sticker pack
        """
        packs = await self.http.get("sticker-packs")
        for pack in packs:
            yield StickerPack.from_dict(pack)

    async def create_stage(
        self,
        channel_id: int,
        topic: str,
        privacy_level: Optional[PrivacyLevel] = None,
        reason: Optional[str] = None,
    ) -> StageInstance:
        """|coro|

        Parameters
        ----------
        channel_id : :class:`int`
            The id of the Stage channel
        topic : :class:`str`
            The topic of the Stage instance (1-120 characters)
        privacy_level : Optional[:class:`~pincer.objects.guild.stage.PrivacyLevel`]
            The privacy level of the Stage instance (default :data:`GUILD_ONLY`)
        reason : Optional[:class:`str`]
            The reason for creating the Stage instance

        Returns
        -------
        :class:`~pincer.objects.guild.stage.StageInstance`
            The Stage instance created
        """

        data = {
            "channel_id": channel_id,
            "topic": topic,
            "privacy_level": privacy_level,
        }

        return await self.http.post(  # type: ignore
            "stage-instances", remove_none(data), headers={"reason": reason}
        )

    async def get_stage(self, _id: int) -> StageInstance:
        """|coro|
        Gets the stage instance associated with the Stage channel, if it exists

        Parameters
        ----------
        _id : int
            The ID of the stage to get

        Returns
        -------
        :class:`~pincer.objects.guild.stage.StageInstance`
            The stage instance
        """
        return await StageInstance.from_id(self, _id)

    async def modify_stage(
        self,
        _id: int,
        topic: Optional[str] = None,
        privacy_level: Optional[PrivacyLevel] = None,
        reason: Optional[str] = None,
    ):
        """|coro|
        Updates fields of an existing Stage instance.
        Requires the user to be a moderator of the Stage channel.

        Parameters
        ----------
        _id : int
            The ID of the stage to modify
        topic : Optional[:class:`str`]
            The topic of the Stage instance (1-120 characters)
        privacy_level : Optional[:class:`~pincer.objects.guild.stage.PrivacyLevel`]
            The privacy level of the Stage instance
        reason : Optional[:class:`str`]
            The reason for the modification
        """

        await self.http.patch(
            f"stage-instances/{_id}",
            remove_none({"topic": topic, "privacy_level": privacy_level}),
            headers={"reason": reason},
        )

    async def delete_stage(self, _id: int, reason: Optional[str] = None):
        """|coro|
        Deletes the Stage instance.
        Requires the user to be a moderator of the Stage channel.

        Parameters
        ----------
        _id : int
            The ID of the stage to delete
        reason : Optional[:class:`str`]
            The reason for the deletion
        """
        await self.http.delete(
            f"stage-instances/{_id}", headers={"reason": reason}
        )

    async def crosspost_message(
        self, channel_id: int, message_id: int
    ) -> UserMessage:
        """|coro|
        Crosspost a message in a News Channel to following channels.

        This endpoint requires the ``SEND_MESSAGES`` permission,
        if the current user sent the message, or additionally the
        ``MANAGE_MESSAGES`` permission, for all other messages,
        to be present for the current user.

        Parameters
        ----------
        channel_id : int
            ID of the news channel that the message is in.
        message_id : int
            ID of the message to crosspost.

        Returns
        -------
        :class:`~pincer.objects.message.UserMessage`
            The crossposted message
        """

        return await self._http.post(
            f"channels/{channel_id}/{message_id}/crosspost"
        )


Bot = Client



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from enum import IntEnum
from typing import overload, TYPE_CHECKING

from aiohttp import FormData

from .channel import Channel, Thread
from .scheduled_events import ScheduledEvent, GuildScheduledEventUser
from ..message.emoji import Emoji
from ..message.file import File
from ...exceptions import UnavailableGuildError
from ...utils import remove_none
from ...utils.api_data import APIDataGen
from ...utils.api_object import APIObject
from ...utils.types import MISSING

from .audit_log import AuditLog
from .ban import Ban
from .member import GuildMember
from .invite import Invite
from .role import Role
from .template import GuildTemplate
from .welcome_screen import WelcomeScreen
from .widget import GuildWidget
from .webhook import Webhook
from ..user.integration import Integration
from ..voice.region import VoiceRegion
from ..message.sticker import Sticker

if TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Tuple, Union, Generator

    from ..events.presence import PresenceUpdateEvent
    from .channel import ChannelType
    from .features import GuildFeature
    from .overwrite import Overwrite
    from .stage import StageInstance
    from .welcome_screen import WelcomeScreenChannel
    from ..user.user import User
    from ..user.voice_state import VoiceState
    from ...client import Client
    from ...utils.timestamp import Timestamp
    from ...utils.types import APINullable, JSONSerializable
    from ...utils.snowflake import Snowflake


class PremiumTier(IntEnum):
    """Represents the boost tier of a guild.
    Attributes
    ----------
    NONE:
        Guild has not unlocked any Server Boost perks.
    TIER_1:
        Guild has unlocked Server Boost level 1 perks.
    TIER_2:
        Guild has unlocked Server Boost level 2 perks.
    TIER_3:
        Guild has unlocked Server Boost level 3 perks.
    """

    NONE = 0
    TIER_1 = 1
    TIER_2 = 2
    TIER_3 = 3


class GuildNSFWLevel(IntEnum):
    """Represents the NSFW level of a guild.
    Attributes
    ----------
    DEFAULT:
        Default NSFW level.
    EXPLICIT:
        Explicit NSFW level.
    SAFE:
        SAFE NSFW level.
    AGE_RESTRICTED:
        Age restricted NSFW level.
    """

    DEFAULT = 0
    EXPLICIT = 1
    SAFE = 2
    AGE_RESTRICTED = 3


class ExplicitContentFilterLevel(IntEnum):
    """Represents the filter content level of a guild.
    Attributes
    ----------
    DISABLED:
        Media content will not be scanned.
    MEMBERS_WITHOUT_ROLES:
        Media content sent by members without roles will be scanned.
    ALL_MEMBERS:
        Media content sent by all members will be scanned.
    """

    DISABLED = 0
    MEMBERS_WITHOUT_ROLES = 1
    ALL_MEMBERS = 2


class MFALevel(IntEnum):
    """Represents the multi-factor authentication level of a guild.
    Attributes
    ----------
    NONE:
        Guild has no MFA/2FA requirement for moderation actions.
    ELEVATED:
        Guild has a 2FA requirement for moderation actions
    """

    NONE = 0
    ELEVATED = 1


class VerificationLevel(IntEnum):
    """Represents the verification level of a guild.
    Attributes
    ----------
    NONE:
        Unrestricted.
    LOW:
        Must have verified email on account.
    MEDIUM:
        Must be registered on Discord for longer than 5 minutes.
    HIGH:
        Must be a member of the server for longer than 10 minutes.
    VERY_HIGH:
        Must have a verified phone number.
    """

    NONE = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3
    VERY_HIGH = 4


class DefaultMessageNotificationLevel(IntEnum):
    """Represents the default message notification level of a guild.
    Attributes
    ----------
    ALL_MESSAGES:
        Members will receive notifications for all messages by default.
    ONLY_MENTIONS:
        Members will receive notifications only for messages that @mention them by default.
    """

    # noqa: E501
    ALL_MESSAGES = 0
    ONLY_MENTIONS = 1


class SystemChannelFlags(IntEnum):
    """Represents the system channel flags of a guild.
    Attributes
    ----------
    SUPPRESS_JOIN_NOTIFICATIONS:
        Suppress member join notifications.
    SUPPRESS_PREMIUM_SUBSCRIPTIONS:
        Suppress server boost notifications.
    SUPPRESS_GUILD_REMINDER_NOTIFICATIONS:
        Suppress server setup tips.
    SUPPRESS_JOIN_NOTIFICATION_REPLIES:
        Hide member join sticker reply buttons
    """

    SUPPRESS_JOIN_NOTIFICATIONS = 1 << 0
    SUPPRESS_PREMIUM_SUBSCRIPTIONS = 1 << 1
    SUPPRESS_GUILD_REMINDER_NOTIFICATIONS = 1 << 2
    SUPPRESS_JOIN_NOTIFICATION_REPLIES = 1 << 3


@dataclass(repr=False)
class GuildPreview(APIObject):
    """Represents a guild preview.
    Attributes
    ----------
    id: :class:`Snowflake`
        The guild ID.
    name: :class:`str`
        The guild name.
    icon: :class:`str`
        The guild icon hash.
    splash: :class:`str`
        The guild splash hash.
    discovery_splash: :class:`str`
        The guild discovery splash hash.
    emojis: :class:`List[Emoji]`
        The guild emojis.
    features: :class:`List[GuildFeature]`
        The guild features.
    approximate_member_count: :class:`int`
        The approximate member count.
    approximate_presence_count: :class:`int`
        The approximate number of online members in this guild
    description: :class:`str`
        The guild description.
    """

    id: Snowflake
    name: str
    emojis: List[Emoji]
    features: List[GuildFeature]
    approximate_member_count: int
    approximate_presence_count: int

    icon: APINullable[str] = MISSING
    splash: APINullable[str] = MISSING
    discovery_splash: APINullable[str] = MISSING
    description: APINullable[str] = MISSING


@dataclass(repr=False)
class Guild(APIObject):
    """Represents a Discord guild/server in which your client resides.
    Attributes
    ----------
    afk_channel_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
        Id of afk channel
    afk_timeout: :class:`int`
        Afk timeout in seconds
    application_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
        Application id of the guild creator if it is bot-created
    banner: Optional[:class:`str`]
        Banner hash
    default_message_notifications: :class:`~pincer.objects.guild.guild.DefaultMessageNotificationLevel`
        Default message notifications level
    description: Optional[:class:`str`]
        The description of a Community guild
    discovery_splash: Optional[:class:`str`]
        Discovery splash hash;
        only present for guilds with the "DISCOVERABLE" feature
    emojis: List[:class:`~pincer.objects.message.emoji.Emoji`]
        Custom guild emojis
    explicit_content_filter: :class:`~pincer.objects.guild.guild.ExplicitContentFilterLevel`
        Explicit content filter level
    features: List[:class:`~pincer.objects.guild.features.GuildFeature`]
        Enabled guild features
    id: :class:`~pincer.utils.snowflake.Snowflake`
        Guild id
    icon: Optional[:class:`str`]
        Icon hash
    mfa_level: :class:`~pincer.objects.guild.guild.MFALevel`
        Required MFA level for the guild
    name: :class:`str`
        Guild name (2-100 characters, excluding trailing and leading
        whitespace)
    nsfw_level: :class:`~pincer.objects.guild.guild.NSFWLevel`
        Guild NSFW level
    owner_id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of owner
    preferred_locale: :class:`str`
        The preferred locale of a Community guild;
        used in server discovery and notices from Discord;
        defaults to "en-US"
    premium_tier: :class:`~pincer.objects.guild.guild.PremiumTier`
        Premium tier (Server Boost level)
    public_updates_channel_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
        The id of the channel where admins
        and moderators of Community guilds receive notices from Discord
    roles: List[:class:`~pincer.objects.guild.role.Role`]
        Roles in the guild
    rules_channel_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
        The id of the channel where Community guilds can display rules
        and/or guidelines
    splash: Optional[:class:`str`]
        Splash hash
    system_channel_flags: :class:`~pincer.objects.guild.guild.SystemChannelFlags`
        System channel flags
    system_channel_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
        The id of the channel where guild notices
        such as welcome messages and boost events are posted
    vanity_url_code: Optional[:class:`str`]
        The vanity url code for the guild
    verification_level: :class:`~pincer.objects.guild.guild.VerificationLevel`
        Verification level required for the guild
    approximate_member_count: APINullable[:class:`int`]
        Approximate number of members in this guild, returned from the
        `GET /guilds/<id>` endpoint when with_counts is true
    approximate_presence_count: APINullable[:class:`int`]
        Approximate number of non-offline members in this guild,
        returned from the `GET /guilds/<id>`
        endpoint when with_counts is true
    channels: APINullable[List[:class:`~pincer.objects.guild.channel.Channel`]]
        Channels in the guild
    icon_hash: APINullable[Optional[:class:`str`]]
        Icon hash, returned when in the template object
    joined_at: APINullable[:class:`~pincer.utils.timestamp.Timestamp`]
        When this guild was joined at
    large: APINullable[:class:`bool`]
        True if this is considered a large guild
    max_members: APINullable[:class:`int`]
        The maximum number of members for the guild
    max_presences: APINullable[Optional[:class:`int`]]
        The maximum number of presences for the guild
        (null is always returned, apart from the largest of guilds)
    max_video_channel_users: APINullable[:class:`int`]
        The maximum amount of users in a video channel
    members: APINullable[List[:class:`~pincer.objects.guild.member.GuildMember`]]
        Users in the guild
    member_count: APINullable[:class:`bool`]
        Total number of members in this guild
    nsfw: APINullable[:class:`bool`]
        Boolean if the server is NSFW
    owner: APINullable[:class:`bool`]
        True if the user is the owner of the guild
    permissions: APINullable[:class:`str`]
        Total permissions for the user in the guild
        (excludes overwrites)
    premium_subscription_count: APINullable[:class:`int`]
        The number of boosts this guild currently has
    presences: APINullable[List[:class:`~pincer.objects.events.presence.PresenceUpdateEvent`]]
        Presences of the members in the guild,
        will only include non-offline members if the size is greater
        than large threshold
    stage_instances: APINullable[List[:class:`~pincer.objects.guild.stage.StageInstance`]]
        Stage instances in the guild
    stickers: Optional[List[:class:`~pincer.objects.message.sticker.Sticker`]]
        Custom guild stickers
    region: APINullable[Optional[:class:`str`]]
        Voice region id for the guild (deprecated)
    threads: APINullable[List[:class:`~pincer.objects.guild.channel.Channel`]]
        All active threads in the guild that current user
        has permission to view
    unavailable: APINullable[:class:`bool`]
        True if this guild is unavailable due to an outage
    voice_states: APINullable[List[:class:`~pincer.objects.user.voice_state.VoiceState`]]
        States of members currently in voice channels;
        lacks the guild_id key
    widget_enabled: APINullable[:class:`bool`]
        True if the server widget is enabled
    widget_channel_id: APINullable[Optional[:class:`~pincer.utils.snowflake.Snowflake`]]
        The channel id that the widget will generate an invite to,
        or null if set to no invite
    welcome_screen: APINullable[:class:`~pincer.objects.guild.welcome_screen.WelcomeScreen`]
        The welcome screen of a Community guild, shown to new members,
        returned in an Invite's guild object
    """

    # noqa: E501
    features: List[GuildFeature]
    id: Snowflake
    name: str
    nsfw_level: GuildNSFWLevel
    verification_level: VerificationLevel

    # Guild invites missing
    system_channel_flags: APINullable[SystemChannelFlags] = MISSING
    explicit_content_filter: APINullable[ExplicitContentFilterLevel] = MISSING
    premium_tier: APINullable[PremiumTier] = MISSING
    default_message_notifications: APINullable[
        DefaultMessageNotificationLevel
    ] = MISSING
    mfa_level: APINullable[MFALevel] = MISSING
    owner_id: APINullable[Snowflake] = MISSING
    afk_timeout: APINullable[int] = MISSING
    emojis: APINullable[List[Emoji]] = MISSING
    preferred_locale: APINullable[str] = MISSING
    roles: APINullable[List[Role]] = MISSING

    guild_scheduled_events: APINullable[List[ScheduledEvent]] = MISSING
    lazy: APINullable[bool] = MISSING
    premium_progress_bar_enabled: APINullable[bool] = MISSING
    guild_hashes: APINullable[Dict] = MISSING
    afk_channel_id: APINullable[Snowflake] = MISSING
    application_id: APINullable[Snowflake] = MISSING
    embedded_activities: APINullable[List] = MISSING
    banner: APINullable[str] = MISSING
    description: APINullable[str] = MISSING
    discovery_splash: APINullable[str] = MISSING
    icon: APINullable[str] = MISSING
    public_updates_channel_id: APINullable[Snowflake] = MISSING
    rules_channel_id: APINullable[Snowflake] = MISSING
    splash: APINullable[str] = MISSING
    system_channel_id: APINullable[Snowflake] = MISSING
    vanity_url_code: APINullable[str] = MISSING

    application_command_counts: APINullable[Dict] = MISSING
    application_command_count: APINullable[int] = MISSING
    approximate_member_count: APINullable[int] = MISSING
    approximate_presence_count: APINullable[int] = MISSING
    channels: APINullable[List[Channel]] = field(default_factory=list)
    # TODO: Add type when type is known
    hub_type: APINullable[Any] = MISSING
    icon_hash: APINullable[Optional[str]] = MISSING
    joined_at: APINullable[Timestamp] = MISSING
    large: APINullable[bool] = MISSING
    max_members: APINullable[int] = MISSING
    max_presences: APINullable[Optional[int]] = MISSING
    max_video_channel_users: APINullable[int] = MISSING
    members: APINullable[List[GuildMember]] = MISSING
    member_count: APINullable[bool] = MISSING
    nsfw: APINullable[bool] = MISSING
    # Note: This is missing from discord's docs but in the api
    owner: APINullable[bool] = MISSING
    permissions: APINullable[str] = MISSING
    premium_subscription_count: APINullable[int] = MISSING
    presences: APINullable[List[PresenceUpdateEvent]] = MISSING
    stage_instances: APINullable[List[StageInstance]] = MISSING
    stickers: APINullable[List[Sticker]] = MISSING
    region: APINullable[Optional[str]] = MISSING
    threads: APINullable[List[Channel]] = MISSING
    # Guilds are considered available unless otherwise specified
    unavailable: APINullable[bool] = False
    voice_states: APINullable[List[VoiceState]] = MISSING
    widget_enabled: APINullable[bool] = MISSING
    widget_channel_id: APINullable[Optional[Snowflake]] = MISSING
    welcome_screen: APINullable[WelcomeScreen] = MISSING

    @classmethod
    async def from_id(
        cls,
        client: Client,
        _id: Union[int, Snowflake],
        with_counts: bool = False,
    ) -> Guild:
        """
        Parameters
        ----------
        client : :class:`~pincer.Client`
            Client object to use the http gateway from.
        _id : :class:`~pincer.utils.snowflake.Snowflake`
            Guild ID.
        Returns
        -------
        :class:`~pincer.objects.guild.guild.Guild`
            The new guild object.
        """
        data = await client.http.get(
            f"/guilds/{_id}",
            # Yarl don't support boolean params
            params={"with_counts": "true" if with_counts else None},
        )
        channel_data = await client.http.get(f"/guilds/{_id}/channels")

        data["channels"]: List[Channel] = [
            Channel.from_dict(i) for i in (channel_data or [])
        ]

        return Guild.from_dict(data)

    async def get_member(self, _id: int) -> GuildMember:
        """|coro|
        Fetches a GuildMember from its identifier

        Parameters
        ----------
        _id: int
            The id of the guild member which should be fetched from the Discord
            gateway.
        Returns
        -------
        :class:`~pincer.objects.guild.member.GuildMember`
            A GuildMember object.
        """
        return await GuildMember.from_id(self._client, self.id, _id)

    @overload
    async def modify_member(
        self,
        *,
        _id: int,
        nick: Optional[str] = None,
        roles: Optional[List[Snowflake]] = None,
        mute: Optional[bool] = None,
        deaf: Optional[bool] = None,
        channel_id: Optional[Snowflake] = None,
        reason: Optional[str] = None,
        communication_disabled_until: Optional[Timestamp] = MISSING,
    ) -> GuildMember:
        """|coro|
        Modifies a member in the guild from its identifier and based on the
        keyword arguments provided.
        Parameters
        ----------
        _id : int
            Id of the member to modify
        nick : Optional[:class:`str`]
            New nickname for the member |default| :data:`None`
        roles : Optional[List[:class:`~pincer.utils.snowflake.Snowflake]]
            New roles for the member |default| :data:`None`
        mute : Optional[:class:`bool`]
            Whether the member is muted |default| :data:`None`
        deaf : Optional[:class:`bool`]
            Whether the member is deafened |default| :data:`None`
        channel_id : Optional[:class:`~pincer.utils.snowflake.Snowflake]
            Voice channel id to move to |default| :data:`None`
        reason : Optional[:class:`str`]
            audit log reason |default| :data:`None`
        communication_disabled_until : Optional[Timestamp]
            When the member can communicate again, requires ``MODERATE_MEMBERS``
            permissions. Set to ``None`` to disable the timeout.

        Returns
        -------
        :class:`~pincer.objects.guild.member.GuildMember`
            The new member object.
        """
        ...

    async def modify_member(
        self, _id: int, reason=None, **kwargs
    ) -> GuildMember:
        if kwargs.get("communication_disabled_until") is MISSING:
            kwargs.pop("communication_disabled_until")
        data = await self._http.patch(
            f"guilds/{self.id}/members/{_id}",
            data=kwargs,
            headers={"X-Audit-Log-Reason": reason},
        )
        return GuildMember.from_dict(data)

    @overload
    async def create_channel(
        self,
        *,
        name: str,
        type: Optional[ChannelType] = None,
        topic: Optional[str] = None,
        bitrate: Optional[int] = None,
        user_limit: Optional[int] = None,
        rate_limit_per_user: Optional[int] = None,
        position: Optional[int] = None,
        permission_overwrites: Optional[List[Overwrite]] = None,
        parent_id: Optional[Snowflake] = None,
        nsfw: Optional[bool] = None,
    ) -> Channel:
        """|coro|
        Create a new channel object for the guild.

        Parameters
        ----------
        name : str
            channel name (1-100 characters)
        type : Optional[:class:int`]
            the type of channel
        topic : Optional[:class:str`]
            channel topic (0-1024 characters)
        bitrate : Optional[:class:`int`]
            the bitrate (in bits) of the voice channel (voice only)
        user_limit : Optional[:class:`int`]
            the user limit of the voice channel (voice only)
        rate_limit_per_user : Optional[:class:`int`]
            amount of seconds a user has to wait
            before sending another message (0-21600)
            bots, as well as users with the permission
            manage_messages or manage_channel, are unaffected
        position : Optional[:class:`int`]
            sorting position of the channel
        permission_overwrites : Optional[List[:class:`~pincer.objects.guild.overwrite.Overwrite`]]
            the channel's permission overwrites
        parent_id : Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            id of the parent category for a channel
        nsfw : Optional[:class:`bool`]
            whether the channel is nsfw
        reason : Optional[:class:`str`]
            audit log reason |default| :data:`None`
        Returns
        -------
        :class:`~pincer.objects.guild.channel.Channel`
            The new channel object.
        """
        ...

    async def create_channel(self, *, reason: Optional[str] = None, **kwargs):
        data = await self._http.post(
            f"guilds/{self.id}/channels",
            data=kwargs,
            headers={"X-Audit-Log-Reason": reason},
        )
        return Channel.from_dict(data)

    async def modify_channel_positions(
        self,
        reason: Optional[str] = None,
        *channel: Dict[str, Optional[Union[int, bool, Snowflake]]],
    ):
        """|coro|
        Create a new channel object for the guild.

        Parameters
        ----------
        reason : Optional[:class:`str`]
            audit log reason |default| :data:`None`
        \\*channel : Dict[str, Optional[Union[int, bool, :class:`~pincer.utils.snowflake.Snowflake`]
            Keys:
                - id : :class:`~pincer.utils.snowflake.Snowflake`
                - position : Optional[:class:`int`]
                - lock_permissions : Optional[:class:`bool`]
                - parent_id : Optional[:class:`~pincer.utils.snowflake.Snowflake`]
        """
        await self._http.patch(
            f"guilds/{self.id}/channels",
            data=channel,
            headers={"X-Audit-Log-Reason": reason},
        )

    async def list_active_threads(
        self,
    ) -> Tuple[Generator[Thread], Generator[GuildMember]]:
        """|coro|
        Returns all active threads in the guild,
        including public and private threads.

        Returns
        -------
        Generator[Union[:class:`~pincer.objects.guild.channel.PublicThread`, :class:`~pincer.objects.guild.channel.PrivateThread`]], Generator[:class:`~pincer.objects.guild.member.GuildMember`]]
            The new member object.
        """
        data = await self._http.get(f"guilds/{self.id}/threads/active")

        threads = (Channel.from_dict(channel) for channel in data["threads"])
        members = (GuildMember.from_dict(member) for member in data["members"])

        return threads, members

    def list_guild_members(
        self, limit: int = 1, after: int = 0
    ) -> APIDataGen[GuildMember]:
        """|coro|
        Returns a list of guild member objects that are members of the guild.

        Parameters
        ----------
        limit : int
            max number of members to return (1-1000) |default| :data:`1`
        after : int
            the highest user id in the previous page |default| :data:`0`

        Yields
        ------
        :class:`~pincer.objects.guild.member.GuildMember`
            the guild member object that is in the guild
        """

        return APIDataGen(
            GuildMember,
            self._http.get(
                f"guilds/{self.id}/members",
                params={"limit": limit, "after": after},
            ),
        )

    def search_guild_members(
        self, query: str, limit: Optional[int] = None
    ) -> APIDataGen[GuildMember]:
        """|coro|
        Returns a list of guild member objects whose
        username or nickname starts with a provided string.

        Parameters
        ----------
        query : str
            Query string to match username(s) and nickname(s) against.
        limit : Optional[int]
            max number of members to return (1-1000) |default| :data:`1`

        Yields
        -------
        :class:`~pincer.objects.guild.member.GuildMember`
            guild member objects
        """

        return APIDataGen(
            GuildMember,
            self._http.get(
                f"guilds/{self.id}/members/search",
                params={"query": query, "limit": limit},
            ),
        )

    @overload
    async def add_guild_member(
        self,
        *,
        user_id: Snowflake,
        access_token: str,
        nick: Optional[str] = None,
        roles: Optional[List[Snowflake]] = None,
        mute: Optional[bool] = None,
        deaf: Optional[bool] = None,
        reason: Optional[str] = None,
    ) -> Optional[GuildMember]:
        """|coro|
        Adds a user to the guild, provided you have a
        valid oauth2 access token for the user with the guilds.join scope.

        Parameters
        ----------
        user_id : str
            id of the user to be added
        access_token : str
            an oauth2 access token granted with the guilds.join to
            the bot's application for the user you want to add to the guild
        nick : Optional[str]
            value to set users nickname to
        roles : Optional[List[:class:`~pincer.utils.snowflake.Snowflake`]]
            array of role ids the member is assigned
        mute : Optional[bool]
            whether the user is muted in voice channels
        deaf : Optional[bool]
            whether the user is deafened in voice channels
        reason : Optional[:class:`str`]
            audit log reason |default| :data:`None`
        Returns
        -------
        :class:`~pincer.objects.guild.member.GuildMember`
            If the user is not in the guild
        None
            If the user is in the guild
        """

    async def add_guild_member(self, user_id, reason=None, **kwargs):
        data = await self._http.put(
            f"guilds/{self.id}/members/{user_id}",
            data=kwargs,
            headers={"X-Audit-Log-Reason": reason},
        )

        return GuildMember.from_dict(data) if data else None

    async def modify_current_member(
        self, nick: str, reason: Optional[str] = None
    ) -> GuildMember:
        """|coro|
        Modifies the current member in a guild.

        Parameters
        ----------
        nick : str
            value to set users nickname to
        reason : Optional[:class:`str`]
            audit log reason |default| :data:`None`
        Returns
        -------
        class:`~pincer.objects.guild.member.GuildMember
            current guild member
        """
        data = self._http.patch(
            f"guilds/{self.id}/members/@me",
            {"nick": nick},
            headers={"X-Audit-Log-Reason": reason},
        )
        return GuildMember.from_dict(data)

    async def add_guild_member_role(
        self, user_id: int, role_id: int, reason: Optional[str] = None
    ) -> None:
        """|coro|
        Adds a role to a guild member.

        Parameters
        ----------
        user_id : int
            id of the user to give a role to
        role_id : int
            id of a role
        reason : Optional[:class:`str`]
            audit log reason |default| :data:`None`
        """
        await self._http.put(
            f"guilds/{self.id}/members/{user_id}/roles/{role_id}",
            headers={"X-Audit-Log-Reason": reason},
        )

    async def remove_guild_member_role(
        self, user_id: int, role_id: int, reason: Optional[str] = None
    ):
        """|coro|
        Removes a role from a guild member.

        Parameters
        ----------
        user_id : int
            id of the user to remove a role from
        role_id : int
            id of a role
        reason : Optional[:class:`str`]
            audit log reason |default| :data:`None`
        """
        await self._http.delete(
            f"guilds/{self.id}/members/{user_id}/roles/{role_id}",
            headers={"X-Audit-Log-Reason": reason},
        )

    async def remove_guild_member(
        self, user_id: int, reason: Optional[str] = None
    ):
        """|coro|
        Remove a member from a guild.

        Parameters
        ----------
        user_id : int
            id of the user to remove from the guild
        reason : Optional[:class:`str`]
            audit log reason |default| :data:`None`
        """
        await self._http.delete(
            f"guilds/{self.id}/members/{user_id}",
            headers={"X-Audit-Log-Reason": reason},
        )

    async def ban(
        self,
        member: Union[int, GuildMember],
        reason: str = None,
        delete_message_days: int = None,
    ):
        """
        Ban a guild member.

        Parameters
        ----------
        member : Union[:class:`int`, :class:`GuildMember`]
            ID or object of the guild member to ban.
        reason : Optional[:class:`str`]
            Reason for the kick.
        delete_message_days : Optional[:class:`int`]
            Number of days to delete messages for (0-7)
        """
        headers = {}
        member_id: int = member if isinstance(member, int) else member.id

        if reason is not None:
            headers["X-Audit-Log-Reason"] = reason

        data = {}

        if delete_message_days is not None:
            data["delete_message_days"] = delete_message_days

        await self._http.put(
            f"/guilds/{self.id}/bans/{member_id}", data=data, headers=headers
        )

    async def kick(
        self, member: Union[int, GuildMember], reason: Optional[str] = None
    ):
        """|coro|
        Kicks a guild member.

        Parameters
        ----------
        member : Union[:class:`int`, :class:`GuildMember`]
            ID or object of the guild member to kick.
        reason : Optional[:class:`str`]
            Reason for the kick.
        """
        headers = {}
        member_id: int = member if isinstance(member, int) else member.id

        if reason is not None:
            headers["X-Audit-Log-Reason"] = reason

        await self._http.delete(
            f"/guilds/{self.id}/members/{member_id}", headers=headers
        )

    def get_roles(self) -> APIDataGen[Role]:
        """|coro|
        Fetches all the roles in the guild.

        Yields
        -------
        AsyncGenerator[:class:`~pincer.objects.guild.role.Role`, :data:`None`]
            An async generator of Role objects.
        """

        return APIDataGen(Role, self._http.get(f"guilds/{self.id}/roles"))

    @overload
    async def create_role(
        self,
        reason: Optional[str] = None,
        *,
        name: Optional[str] = "new role",
        permissions: Optional[str] = None,
        color: Optional[int] = 0,
        hoist: Optional[bool] = False,
        icon: Optional[str] = None,
        unicode_emoji: Optional[str] = None,
        mentionable: Optional[bool] = False,
    ) -> Role:
        """|coro|
        Creates a new role for the guild.
        Requires the ``MANAGE_ROLES`` permission.

        Parameters
        ----------
        reason : Optional[:class:`str`]
            Reason for creating the role. |default| :data:`None`
        name : Optional[:class:`str`]
            name of the role |default| :data:`"new role"`
        permissions : Optional[:class:`str`]
            bitwise value of the enabled/disabled
            permissions, set to @everyone permissions
            by default |default| :data:`None`
        color : Optional[:class:`int`]
            RGB color value |default| :data:`0`
        hoist : Optional[:class:`bool`]
            whether the role should be displayed
            separately in the sidebar |default| :data:`False`
        icon : Optional[:class:`str`]
            the role's icon image (if the guild has
            the ``ROLE_ICONS`` feature) |default| :data:`None`
        unicode_emoji : Optional[:class:`str`]
            the role's unicode emoji as a standard emoji (if the guild
            has the ``ROLE_ICONS`` feature) |default| :data:`None`
        mentionable : Optional[:class:`bool`]
            whether the role should be mentionable |default| :data:`False`

        Returns
        -------
        :class:`~pincer.objects.guild.role.Role`
            The new role object.
        """
        ...

    async def create_role(self, reason: Optional[str] = None, **kwargs) -> Role:
        return Role.from_dict(
            await self._http.post(
                f"guilds/{self.id}/roles",
                data=kwargs,
                headers={"X-Audit-Log-Reason": reason},
            )
        )

    def edit_role_position(
        self,
        id: Snowflake,
        reason: Optional[str] = None,
        position: Optional[int] = None,
    ) -> APIDataGen[Role]:
        """|coro|
        Edits the position of a role.

        Parameters
        ----------
        id : :class:`~pincer.utils.snowflake.Snowflake`
            The role ID
        reason : Optional[:class:`str`]
            Reason for editing the role position. |default| :data:`None`
        position : Optional[:class:`int`]
            Sorting position of the role |default| :data:`None`

        Yields
        -------
        AsyncGenerator[:class:`~pincer.objects.guild.role.Role`, :data:`None`]
            An async generator of all the guild's role objects.
        """
        return APIDataGen(
            Role,
            self._http.patch(
                f"guilds/{self.id}/roles",
                data={"id": id, "position": position},
                headers={"X-Audit-Log-Reason": reason},
            ),
        )

    @overload
    async def edit_role(
        self,
        id: Snowflake,
        reason: Optional[str] = None,
        *,
        name: Optional[str] = None,
        permissions: Optional[str] = None,
        color: Optional[int] = None,
        hoist: Optional[bool] = None,
        icon: Optional[str] = None,
        unicode_emoji: Optional[str] = None,
        mentionable: Optional[bool] = None,
    ) -> Role:
        """|coro|
        Edits a role.
        Requires the ``MANAGE_ROLES`` permission.

        Parameters
        ----------
        id : :class:`~pincer.utils.snowflake.Snowflake`
            The role ID
        reason : Optional[:class:`str`]
            Reason for editing the role |default| :data:`None`
        name : Optional[:class:`str`]
            Name of the role |default| :data:`None`
        permissions : Optional[:class:`str`]
            Bitwise value of the enabled/disabled
            permissions |default| :data:`None`
        color : Optional[:class:`int`]
            RGB color value |default| :data:`None`
        hoist : Optional[:class:`bool`]
            Whether the role should be displayed
            separately in the sidebar |default| :data:`None`
        icon : Optional[:class:`str`]
            The role's icon image (if the guild has
            the ``ROLE_ICONS`` feature) |default| :data:`None`
        unicode_emoji : Optional[:class:`str`]
            The role's unicode emoji as a standard emoji (if the guild
            has the ``ROLE_ICONS`` feature) |default| :data:`None`
        mentionable : Optional[:class:`bool`]
            Whether the role should be mentionable |default| :data:`None`

        Returns
        -------
        :class:`~pincer.objects.guild.role.Role`
            The edited role object.
        """
        ...

    async def edit_role(
        self, id: Snowflake, reason: Optional[str] = None, **kwargs
    ) -> Role:
        return Role.from_dict(
            await self._http.patch(
                f"guilds/{self.id}/roles/{id}",
                data=kwargs,
                headers={"X-Audit-Log-Reason": reason},
            )
        )

    async def delete_role(self, id: Snowflake, reason: Optional[str] = None):
        """|coro|
        Deletes a role.
        Requires the `MANAGE_ROLES` permission.

        Parameters
        ----------
        id : :class:`~pincer.utils.snowflake.Snowflake`
            The role ID
        reason : Optional[:class:`str`]
            The reason for deleting the role |default| :data:`None`
        """
        await self._http.delete(
            f"guilds/{self.id}/roles/{id}",
            headers={"X-Audit-Log-Reason": reason},
        )

    def get_bans(self) -> APIDataGen[Ban]:
        """|coro|
        Fetches all the bans in the guild.

        Yields
        -------
        AsyncGenerator[:class:`~pincer.objects.guild.ban.Ban`, :data:`None`]
            An async generator of Ban objects.
        """

        return APIDataGen(Ban, self._http.get(f"guilds/{self.id}/bans"))

    async def get_ban(self, id: Snowflake) -> Ban:
        """|coro|
        Fetches a ban from the guild.

        Parameters
        ----------
        id : :class:`~pincer.utils.snowflake.Snowflake`
            The user ID

        Returns
        -------
        :class:`~pincer.objects.guild.ban.Ban`
            The Ban object.
        """
        return Ban.from_dict(
            await self._http.get(f"guilds/{self.id}/bans/{id}")
        )

    async def unban(self, id: Snowflake, reason: Optional[str] = None):
        """|coro|
        Unbans a user from the guild.

        Parameters
        ----------
        id : :class:`~pincer.utils.snowflake.Snowflake`
            The user ID
        reason : Optional[:class:`str`]
            The reason for unbanning the user |default| :data:`None`
        """
        await self._http.delete(
            f"guilds/{self.id}/bans/{id}",
            headers={"X-Audit-Log-Reason": reason},
        )

    @overload
    async def edit(
        self,
        *,
        name: Optional[str] = None,
        region: Optional[str] = None,
        verification_level: Optional[int] = None,
        default_message_notifications: Optional[int] = None,
        explicit_content_filter: Optional[int] = None,
        afk_channel_id: Optional[Snowflake] = None,
        afk_timeout: Optional[int] = None,
        icon: Optional[str] = None,
        owner_id: Optional[Snowflake] = None,
        splash: Optional[str] = None,
        discovery_splash: Optional[str] = None,
        banner: Optional[str] = None,
        system_channel_id: Optional[Snowflake] = None,
        system_channel_flags: Optional[int] = None,
        rules_channel_id: Optional[Snowflake] = None,
        public_updates_channel_id: Optional[Snowflake] = None,
        preferred_locale: Optional[str] = None,
        features: Optional[List[GuildFeature]] = None,
        description: Optional[str] = None,
    ) -> Guild:
        """|coro|
        Modifies the guild

        Parameters
        ----------
        name : Optional[:class:`str`]
            Guild name |default| :data:`None`
        region : Optional[:class:`str`]
            Guild voice region ID |default| :data:`None`
        verification_level : Optional[:class:`int`]
            Verification level |default| :data:`None`
        default_message_notifications : Optional[:class:`int`]
            Default message notification level |default| :data:`None`
        explicit_content_filter : Optional[:class:`int`]
            Explicit content filter level |default| :data:`None`
        afk_channel_id : Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            ID for AFK channel |default| :data:`None`
        afk_timeout : Optional[:class:`int`]
            AFK timeout in seconds |default| :data:`None`
        icon : Optional[:class:`str`]
            base64 1024x1024 png/jpeg/gif image for the guild icon
            (can be animated gif when the server
            has the `ANIMATED_ICON` feature) |default| :data:`None`
        owner_id : Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            User ID to transfer guild ownership to (must be owner) |default| :data:`None`
        splash : Optional[:class:`str`]
            base64 16:9 png/jpeg image for the guild splash (when the
            server has the `INVITE_SPLASH` feature) |default| :data:`None`
        discovery_splash : Optional[:class:`str`]
            base64 16:9 png/jpeg image for the guild discovery splash
            (when the server has the `DISCOVERABLE` feature) |default| :data:`None`
        banner : Optional[:class:`str`]
            base64 16:9 png/jpeg image for the guild banner (when the
            server has the `BANNER` feature) |default| :data:`None`
        system_channel_id : Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            The ID of the channel where guild notices such as welcome
            messages and boost events are posted |default| :data:`None`
        system_channel_flags : Optional[:class:`int`]
            System channel flags |default| :data:`None`
        rules_channel_id : Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            The ID of the channel where Community guilds display rules
            and/or guidelines |default| :data:`None`
        public_updates_channel_id : Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            The ID of the channel where admins and moderators of
            Community guilds receive notices from Discord |default| :data:`None`
        preferred_locale : Optional[:class:`str`]
            The preferred locale of a Community guild used in server
            discovery and notices from Discord; defaults to "en-US" |default| :data:`None`
        features : Optional[List[:class:`GuildFeature`]]
            Enabled guild features |default| :data:`None`
        description : Optional[:class:`str`]
            The description for the guild, if the guild is discoverable |default| :data:`None`

        Returns
        -------
        :class:`~pincer.objects.guild.Guild`
            The modified guild object.
        """
        ...

    async def edit(self, **kwargs) -> Guild:
        g = await self._http.patch(f"guilds/{self.id}", data=kwargs)
        return Guild.from_dict(g)

    async def preview(self) -> GuildPreview:
        """|coro|
        Previews the guild.

        Returns
        -------
        :class:`~pincer.objects.guild.guild.GuildPreview`
            The guild preview object.
        """
        data = await self._http.get(f"guilds/{self.id}/preview")
        return GuildPreview.from_dict(data)

    async def delete(self):
        """|coro|
        Deletes the guild. Returns `204 No Content` on success.
        """
        await self._http.delete(f"guilds/{self.id}")

    async def prune_count(
        self, days: Optional[int] = 7, include_roles: Optional[str] = None
    ) -> int:
        """|coro|
        Returns the number of members that
        would be removed in a prune operation.
        Requires the ``KICK_MEMBERS`` permission.

        Parameters
        ----------
        days : Optional[:class:`int`]
            Number of days to count prune for (1-30) |default| :data:`7`
        include_roles : Optional[:class:`str`]
            Comma-delimited array of Snowflakes;
            role(s) to include |default| :data:`None`

        Returns
        -------
        :class:`int`
            The number of members that would be removed.
        """
        return await self._http.get(
            f"guilds/{self.id}/prune",
            params={"days": days, "include_roles": include_roles},
        )["pruned"]

    async def prune(
        self,
        days: Optional[int] = 7,
        compute_prune_days: Optional[bool] = True,
        include_roles: Optional[List[Snowflake]] = None,
        reason: Optional[str] = None,
    ) -> int:
        """|coro|
        Prunes members from the guild. Requires the ``KICK_MEMBERS`` permission.

        Parameters

        Parameters
        ----------
        days : Optional[:class:`int`]
            Number of days to prune (1-30) |default| :data:`7`
        compute_prune_days : Optional[:class:`bool`]
            Whether ``pruned`` is returned, discouraged for large guilds
            |default| :data:`True`
        include_roles : Optional[List[:class:`~pincer.utils.snowflake.Snowflake`]]
            role(s) to include |default| :data:`None`
        reason : Optional[:class:`str`]
            Reason for the prune |default| :data:`None`

        Returns
        -------
        :class:`int`
            The number of members that were removed.
        """
        return await self._http.post(
            f"guilds/{self.id}/prune",
            data={
                "days": days,
                "compute_prune_days": compute_prune_days,
                "include_roles": include_roles,
            },
            headers={"X-Audit-Log-Reason": reason},
        )["pruned"]

    def get_voice_regions(self) -> APIDataGen[VoiceRegion]:
        """|coro|
        Returns an async generator of voice regions.

        Yields
        -------
        AsyncGenerator[:class:`~pincer.objects.voice.VoiceRegion`, :data:`None`]
            An async generator of voice regions.
        """

        return APIDataGen(
            VoiceRegion, self._http.get(f"guilds/{self.id}/regions")
        )

    def get_invites(self) -> APIDataGen[Invite]:
        """|coro|
        Returns an async generator of invites for the guild.
        Requires the ``MANAGE_GUILD`` permission.

        Yields
        -------
        AsyncGenerator[:class:`~pincer.objects.invite.Invite`, :data:`None`]
            An async generator of invites.
        """

        return APIDataGen(Invite, self._http.get(f"guilds/{self.id}/invites"))

    async def get_invite(self, code: str) -> Invite:
        """|coro|
        Returns an Invite object for the given invite code.

        Parameters
        ----------
        code : :class:`str`
            The invite code to get the invite for.

        Returns
        -------
        :class:`~pincer.objects.guild.invite.Invite`
            The invite object.
        """
        data = await self._http.get(f"invite/{code}")
        return Invite.from_dict(data)

    def get_integrations(self) -> APIDataGen[Integration]:
        """|coro|
        Returns an async generator of integrations for the guild.
        Requires the ``MANAGE_GUILD`` permission.

        Yields
        -------
        AsyncGenerator[:class:`~pincer.objects.integration.Integration`, :data:`None`]
            An async generator of integrations.
        """

        return APIDataGen(
            Integration, self._http.get(f"guilds/{self.id}/integrations")
        )

    async def delete_integration(
        self, integration: Integration, reason: Optional[str] = None
    ):
        """|coro|
        Deletes an integration.
        Requires the ``MANAGE_GUILD`` permission.

        Parameters
        ----------
        integration : :class:`~pincer.objects.integration.Integration`
            The integration to delete.
        reason : Optional[:class:`str`]
            Reason for the deletion |default| :data:`None`
        """
        await self._http.delete(
            f"guilds/{self.id}/integrations/{integration.id}",
            headers={"X-Audit-Log-Reason": reason},
        )

    async def delete_invite(self, code: str):
        """|coro|
        Deletes an invite.
        Requires the ``MANAGE_GUILD`` intent.

        Parameters
        ----------
        code : :class:`str`
            The code of the invite to delete.
        """
        await self._http.delete(f"guilds/{self.id}/invites/{code}")

    async def get_widget_settings(self) -> GuildWidget:
        """|coro|
        Returns the guild widget settings.
        Requires the ``MANAGE_GUILD`` permission.

        Returns
        -------
        :class:`~pincer.objects.guild.widget.GuildWidget`
            The guild widget settings.
        """
        return GuildWidget.from_dict(
            await self._http.get(f"guilds/{self.id}/widget")
        )

    async def modify_widget(
        self, reason: Optional[str] = None, **kwargs
    ) -> GuildWidget:
        """|coro|
        Modifies the guild widget for the guild.
        Requires the ``MANAGE_GUILD`` permission.

        Parameters
        ----------
        reason : Optional[:class:`str`]
            Reason for the modification |default| :data:`None`
        \\*\\*kwargs
            The widget settings to modify

        Returns
        -------
        :class:`~pincer.objects.guild.widget.GuildWidget`
            The updated GuildWidget object
        """
        data = await self._http.patch(
            f"guilds/{self.id}/widget",
            data=kwargs,
            headers={"X-Audit-Log-Reason": reason},
        )
        return GuildWidget.from_dict(data)

    async def get_widget(self) -> Dict[str, JSONSerializable]:
        """|coro|
        Returns the widget for the guild
        """
        return await self._http.get(f"guilds/{self.id}/widget.json")

    @property
    async def vanity_url(self) -> Invite:
        """|coro|
        Returns the Vanity URL for the guild.
        Requires the ``MANAGE_GUILD`` permission.
        ``code`` will be null if a vanity URL has not been set.

        Returns
        -------
        :class:`~pincer.objects.guild.invite.Invite`
            The vanity url for the guild.
        """
        data = await self._http.get(f"guilds/{self.id}/vanity-url")
        return Invite.from_dict(data)

    async def get_widget_image(
        self, style: Optional[str] = "shield"
    ) -> str:  # TODO Replace str with ImageURL object
        """|coro|
        Returns a PNG image widget for the guild.
        Requires no permissions or authentication.

        Widget Style Options
        -------------------
        * [``shield``](https://discord.com/api/guilds/81384788765712384/widget.png?style=shield)
          shield style widget with Discord icon and guild members online count
        * [``banner1``](https://discord.com/api/guilds/81384788765712384/widget.png?style=banner1)
          large image with guild icon, name and online count.
          "POWERED BY DISCORD" as the footer of the widget
        * [``banner2``](https://discord.com/api/guilds/81384788765712384/widget.png?style=banner2)
          smaller widget style with guild icon, name and online count.
          Split on the right with Discord logo
        * [``banner3``](https://discord.com/api/guilds/81384788765712384/widget.png?style=banner3)
          large image with guild icon, name and online count.
          In the footer, Discord logo on the
          left and "Chat Now" on the right
        * [``banner4``](https://discord.com/api/guilds/81384788765712384/widget.png?style=banner4)
          large Discord logo at the top of the widget.
          Guild icon, name and online count in the middle portion
          of the widget and a "JOIN MY SERVER" button at the bottom

        Parameters
        ----------
        style : Optional[:class:`str`]
            Style of the widget image returned |default| :data:`"shield"`

        Returns
        -------
        :class:`str`
            A PNG image of the guild widget.
        """
        return await self._http.get(f"guilds/{self.id}/widget.png?{style=!s}")

    async def get_welcome_screen(self) -> WelcomeScreen:
        """Returns the welcome screen for the guild.

        Returns
        -------
        :class:`~pincer.objects.guild.welcome_screen.WelcomeScreen`
            The welcome screen for the guild.
        """
        data = await self._http.get(f"guilds/{self.id}/welcome-screen")
        return WelcomeScreen.from_dict(data)

    async def modify_welcome_screen(
        self,
        enabled: Optional[bool] = None,
        welcome_channels: Optional[List[WelcomeScreenChannel]] = None,
        description: Optional[str] = None,
        reason: Optional[str] = None,
    ) -> WelcomeScreen:
        """|coro|
        Modifies the guild's Welcome Screen.
        Requires the ``MANAGE_GUILD`` permission.

        Parameters
        ----------
        enabled : Optional[:class:`bool`]
            Whether the welcome screen is enabled |default| :data:`None`
        welcome_channels : Optional[List[:class:`~pincer.objects.guild.welcome_screen.WelcomeScreenChannel`]]
            Channels linked in the welcome screen and
            their display options |default| :data:`None`
        description : Optional[:class:`str`]
            The server description to show
            in the welcome screen |default| :data:`None`
        reason : Optional[:class:`str`]
            Reason for the modification |default| :data:`None`

        Returns
        -------
        :class:`~pincer.objects.guild.welcome_screen.WelcomeScreen`
            The updated WelcomeScreen object
        """
        data = await self._http.patch(
            f"guilds/{self.id}/welcome-screen",
            data={
                "enabled": enabled,
                "welcome_channels": welcome_channels,
                "description": description,
            },
            headers={"X-Audit-Log-Reason": reason},
        )
        return WelcomeScreen.from_dict(data)

    async def modify_current_user_voice_state(
        self,
        channel_id: Snowflake,
        suppress: Optional[bool] = None,
        request_to_speak_timestamp: Optional[Timestamp] = None,
    ):
        """|coro|
        Updates the current user's voice state.

        There are currently several caveats for this endpoint:
        * ``channel_id`` must currently point to a stage channel
        * current user must already have joined ``channel_id``
        * You must have the ``MUTE_MEMBERS`` permission to
          unsuppress yourself. You can always suppress yourself.
        * You must have the ``REQUEST_TO_SPEAK`` permission to request
          to speak. You can always clear your own request to speak.
        * You are able to set ``request_to_speak_timestamp`` to any
          present or future time.

        Parameters
        ----------
        channel_id : :class:`~pincer.utils.snowflake.Snowflake`
            The ID of the channel the user is currently in
        suppress : Optional[:class:`bool`]
            Toggles the user's suppress state |default| :data:`None`
        request_to_speak_timestamp : Optional[:class:`~pincer.utils.timestamp.Timestamp`]
            Sets the user's request to speak
        """
        await self._http.patch(
            f"guilds/{self.id}/voice-states/@me",
            data={
                "channel_id": channel_id,
                "suppress": suppress,
                "request_to_speak_timestamp": request_to_speak_timestamp,
            },
        )

    async def modify_user_voice_state(
        self, user: User, channel_id: Snowflake, suppress: Optional[bool] = None
    ):
        """|coro|
        Updates another user's voice state.

        There are currently several caveats for this endpoint:
        * ``channel_id`` must currently point to a stage channel
        * User must already have joined ``channel_id``
        * You must have the ``MUTE_MEMBERS`` permission.
          (Since suppression is the only thing that is available currently.)
        * When unsuppressed, non-bot users will have their
          ``request_to_speak_timestamp`` set to the current time.
          Bot users will not.
        * When suppressed, the user will have their
          ``request_to_speak_timestamp`` removed.

        Parameters
        ----------
        user : :class:`~pincer.objects.guild.member.Member`
            The user to update
        channel_id : :class:`~pincer.utils.snowflake.Snowflake`
            The ID of the channel the user is currently in
        suppress : Optional[:class:`bool`]
            Toggles the user's suppress state |default| :data:`None`
        """
        await self._http.patch(
            f"guilds/{self.id}/voice-states/{user.id}",
            data={"channel_id": channel_id, "suppress": suppress},
        )

    async def get_audit_log(self) -> AuditLog:
        """|coro|
        Returns an audit log object for the guild.
        Requires the ``VIEW_AUDIT_LOG`` permission.

        Returns
        -------
        :class:`~pincer.objects.guild.audit_log.AuditLog`
            The audit log object for the guild.
        """
        return AuditLog.from_dict(
            await self._http.get(f"guilds/{self.id}/audit-logs")
        )

    def get_emojis(self) -> APIDataGen[Emoji]:
        """|coro|
        Returns an async generator of the emojis in the guild.

        Yields
        ------
        :class:`~pincer.objects.guild.emoji.Emoji`
            The emoji object.
        """
        return APIDataGen(Emoji, self._http.get(f"guilds/{self.id}/emojis"))

    async def get_emoji(self, id: Snowflake) -> Emoji:
        """|coro|
        Returns an emoji object for the given ID.

        Parameters
        ----------
        id : :class:`~pincer.utils.snowflake.Snowflake`
            The ID of the emoji

        Returns
        -------
        :class:`~pincer.objects.guild.emoji.Emoji`
            The emoji object.
        """
        return Emoji.from_dict(
            await self._http.get(f"guilds/{self.id}/emojis/{id}")
        )

    async def create_emoji(
        self,
        *,
        name: str,
        image: File,
        roles: Optional[List[Snowflake]] = None,
        reason: Optional[str] = None,
    ) -> Emoji:
        """|coro|
        Creates a new emoji for the guild.
        Requires the ``MANAGE_EMOJIS_AND_STICKERS`` permission.

        Emojis and animated emojis have a maximum file size of 256kb.
        Attempting to upload an emoji larger than this limit will fail.

        Parameters
        ----------
        name : :class:`str`
            Name of the emoji
        image : :class:`~pincer.objects.message.file.File`
            The File for the 128x128 emoji image data
        roles : Optional[List[:class:`~pincer.utils.snowflake.Snowflake`]]
            Roles allowed to use this emoji |default| :data:`[]`
        reason : Optional[:class:`str`]
            The reason for creating the emoji |default| :data:`None`

        Returns
        -------
        :class:`~pincer.objects.guild.emoji.Emoji`
            The newly created emoji object.
        """
        roles = [] if roles is None else roles

        data = await self._http.post(
            f"guilds/{self.id}/emojis",
            data={"name": name, "image": image.uri, "roles": roles},
            headers={"X-Audit-Log-Reason": reason},
        )
        return Emoji.from_dict(data)

    async def edit_emoji(
        self,
        id: Snowflake,
        *,
        name: Optional[str] = None,
        roles: Optional[List[Snowflake]] = None,
        reason: Optional[str] = None,
    ) -> Emoji:
        """|coro|
        Modifies the given emoji.
        Requires the ``MANAGE_EMOJIS_AND_STICKERS`` permission.

        Parameters
        ----------
        id : :class:`~pincer.utils.snowflake.Snowflake`
            The ID of the emoji
        name : Optional[:class:`str`]
            Name of the emoji |default| :data:`None`
        roles : Optional[List[:class:`~pincer.utils.snowflake.Snowflake`]]
            Roles allowed to use this emoji |default| :data:`None`
        reason : Optional[:class:`str`]
            The reason for editing the emoji |default| :data:`None`

        Returns
        -------
        :class:`~pincer.objects.guild.emoji.Emoji`
            The modified emoji object.
        """
        data = await self._http.patch(
            f"guilds/{self.id}/emojis/{id}",
            data={"name": name, "roles": roles},
            headers={"X-Audit-Log-Reason": reason},
        )
        return Emoji.from_dict(data)

    async def delete_emoji(
        self, id: Snowflake, *, reason: Optional[str] = None
    ):
        """|coro|
        Deletes the given emoji.
        Requires the ``MANAGE_EMOJIS_AND_STICKERS`` permission.

        Parameters
        ----------
        id : :class:`~pincer.utils.snowflake.Snowflake`
            The ID of the emoji
        reason : Optional[:class:`str`]
            The reason for deleting the emoji |default| :data:`None`
        """
        await self._http.delete(
            f"guilds/{self.id}/emojis/{id}",
            headers={"X-Audit-Log-Reason": reason},
        )

    def get_templates(self) -> APIDataGen[GuildTemplate]:
        """|coro|
        Returns an async generator of the guild templates.

        Yields
        -------
        AsyncGenerator[:class:`~pincer.objects.guild.template.GuildTemplate`, :data:`None`]
            The guild template object.
        """

        return APIDataGen(
            GuildTemplate, self._http.get(f"guilds/{self.id}/templates")
        )

    async def create_template(
        self, name: str, description: Optional[str] = None
    ) -> GuildTemplate:
        """|coro|
        Creates a new template for the guild.
        Requires the ``MANAGE_GUILD`` permission.

        Parameters
        ----------
        name : :class:`str`
            Name of the template (1-100 characters)
        description : Optional[:class:`str`]
            Description of the template
            (0-120 characters) |default| :data:`None`
        Returns
        -------
        :class:`~pincer.objects.guild.template.GuildTemplate`
            The newly created template object.
        """
        data = await self._http.post(
            f"guilds/{self.id}/templates",
            data={"name": name, "description": description},
        )
        return GuildTemplate.from_dict(data)

    async def sync_template(self, template: GuildTemplate) -> GuildTemplate:
        """|coro|
        Syncs the given template.
        Requires the ``MANAGE_GUILD`` permission.

        Parameters
        ----------
        template : :class:`~pincer.objects.guild.template.GuildTemplate`
            The template to sync

        Returns
        -------
        :class:`~pincer.objects.guild.template.GuildTemplate`
            The synced template object.
        """
        data = await self._http.put(
            f"guilds/{self.id}/templates/{template.code}"
        )
        return GuildTemplate.from_dict(data)

    async def edit_template(
        self,
        template: GuildTemplate,
        *,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ) -> GuildTemplate:
        """|coro|
        Modifies the template's metadata.
        Requires the ``MANAGE_GUILD`` permission.

        Parameters
        ----------
        template : :class:`~pincer.objects.guild.template.GuildTemplate`
            The template to edit
        name : Optional[:class:`str`]
            Name of the template (1-100 characters)
            |default| :data:`None`
        description : Optional[:class:`str`]
            Description of the template (0-120 characters)
            |default| :data:`None`

        Returns
        -------
        :class:`~pincer.objects.guild.template.GuildTemplate`
            The edited template object.
        """
        data = await self._http.patch(
            f"guilds/{self.id}/templates/{template.code}",
            data={"name": name, "description": description},
        )
        return GuildTemplate.from_dict(data)

    async def delete_template(self, template: GuildTemplate) -> GuildTemplate:
        """|coro|
        Deletes the given template.
        Requires the ``MANAGE_GUILD`` permission.

        Parameters
        ----------
        template : :class:`~pincer.objects.guild.template.GuildTemplate`
            The template to delete

        Returns
        -------
        :class:`~pincer.objects.guild.template.GuildTemplate`
            The deleted template object.
        """
        data = await self._http.delete(
            f"guilds/{self.id}/templates/{template.code}"
        )
        return GuildTemplate.from_dict(data)

    def list_stickers(self) -> APIDataGen[Sticker]:
        """|coro|
        Yields sticker objects for the current guild.
        Includes ``user`` fields if the bot has the
        ``MANAGE_EMOJIS_AND_STICKERS`` permission.

        Yields
        ------
        :class:`~pincer.objects.message.sticker.Sticker`
            a sticker for the current guild
        """

        return APIDataGen(Sticker, self._http.get(f"guild/{self.id}/stickers"))

    async def get_sticker(self, _id: Snowflake) -> Sticker:
        """|coro|
        Returns a sticker object for the current guild and sticker IDs.
        Includes the ``user`` field if the bot has the
        ``MANAGE_EMOJIS_AND_STICKERS`` permission.

        Parameters
        ----------
        _id : int
            id of the sticker

        Returns
        -------
        :class:`~pincer.objects.message.sticker.Sticker`
            the sticker requested
        """
        sticker = await self._http.get(f"guilds/{self.id}/stickers/{_id}")
        return Sticker.from_dict(sticker)

    async def create_sticker(
        self,
        name: str,
        tags: str,
        description: str,
        file: File,
        reason: Optional[str] = None,
    ) -> Sticker:
        """|coro|
        Create a new sticker for the guild.
        Requires the ``MANAGE_EMOJIS_AND_STICKERS permission``.

        Parameters
        ----------
        name : str
            name of the sticker (2-30 characters)
        tags : str
            autocomplete/suggestion tags for the sticker (max 200 characters)
        file : :class:`~pincer.objects.message.file.File`
            the sticker file to upload, must be a PNG, APNG, or Lottie JSON file, max 500 KB
        description : str
            description of the sticker (empty or 2-100 characters) |default| :data:`""`
        reason : Optional[:class:`str`] |default| :data:`None`
            reason for creating the sticker

        Returns
        -------
        :class:`~pincer.objects.message.sticker.Sticker`
            the newly created sticker
        """  # noqa: E501

        form = FormData()
        form.add_field("name", name)
        form.add_field("tags", tags)
        form.add_field("description", description)
        form.add_field("file", file.content, content_type=file.content_type)

        payload = form()

        sticker = await self._http.post(
            f"guilds/{self.id}/stickers",
            data=payload,
            headers={"X-Audit-Log-Reason": reason},
            content_type=payload.content_type,
        )

        return Sticker.from_dict(sticker)

    async def delete_sticker(self, _id: Snowflake):
        """|coro|
        Delete the given sticker.
        Requires the ``MANAGE_EMOJIS_AND_STICKERS`` permission.

        Parameters
        ----------
        _id: Snowflake
            id of the sticker
        """
        await self._http.delete(f"guilds/{self.id}/stickers/{_id}")

    def get_webhooks(self) -> APIDataGen[Webhook]:
        """|coro|
        Returns an async generator of the guild webhooks.

        Yields
        -------
        AsyncGenerator[:class:`~pincer.objects.guild.webhook.Webhook`, None]
            The guild webhook object.
        """

        return APIDataGen(Webhook, self._http.get(f"guilds/{self.id}/webhooks"))

    def get_scheduled_events(
        self, with_user_count: bool = False
    ) -> APIDataGen[ScheduledEvent]:
        """
        Returns an async generator of the guild scheduled events.

        Parameters
        ----------
        with_user_count : :class:`bool`
            Whether to include the user count in the scheduled event.

        Yields
        ------
        :class:`~pincer.objects.guild.scheduled_event.ScheduledEvent`
            The scheduled event object.
        """

        return APIDataGen(
            ScheduledEvent,
            self._http.get(
                f"guilds/{self.id}/scheduled-events",
                param={"with_user_count": with_user_count},
            ),
        )

    async def create_scheduled_event(
        self,
        name: str,
        privacy_level: int,
        entity_type: int,
        scheduled_start_time: datetime,
        scheduled_end_time: Optional[datetime] = None,
        entity_metadata: Optional[str] = None,
        channel_id: Optional[int] = None,
        description: Optional[str] = None,
        reason: Optional[str] = None,
    ) -> ScheduledEvent:
        """
        Create a new scheduled event for the guild.

        Parameters
        ----------
        name : :class:`str`
            The name of the scheduled event.
        privacy_level : :class:`int`
            The privacy level of the scheduled event.
        entity_type : :class:`int`
            The type of entity to be scheduled.
        scheduled_start_time : :class:`datetime`
            The scheduled start time of the event.
        scheduled_end_time : Optional[:class:`datetime`]
            The scheduled end time of the event.
        entity_metadata : Optional[:class:`str`]
            The metadata of the entity to be scheduled.
        channel_id : Optional[:class:`int`]
            The channel id of the channel to be scheduled.
        description : Optional[:class:`str`]
            The description of the scheduled event.
        reason : Optional[:class:`str`]
            The reason for creating the scheduled event.

        Raises
        ------
        ValueError:
            If an event is created in the past or if an event ends before it starts

        Returns
        -------
        :class:`~pincer.objects.guild.scheduled_event.ScheduledEvent`
            The newly created scheduled event.
        """
        if scheduled_start_time < datetime.now():
            raise ValueError("An event cannot be created in the past")

        if scheduled_end_time and scheduled_end_time < scheduled_start_time:
            raise ValueError("An event cannot start before it ends")

        data = await self._http.post(
            f"guilds/{self.id}/scheduled-events",
            data={
                "name": name,
                "privacy_level": privacy_level,
                "entity_type": entity_type,
                "scheduled_start_time": scheduled_start_time.isoformat(),
                "scheduled_end_time": scheduled_end_time.isoformat()
                if scheduled_end_time is not None
                else None,
                "entity_metadata": entity_metadata,
                "channel_id": channel_id,
                "description": description,
            },
            headers={"X-Audit-Log-Reason": reason},
        )
        return ScheduledEvent.from_dict(data)

    async def get_scheduled_event(
        self, _id: int, with_user_count: bool = False
    ) -> ScheduledEvent:
        """
        Get a scheduled event by id.

        Parameters
        ----------
        _id : :class:`int`
            The id of the scheduled event.
        with_user_count : :class:`bool`
            Whether to include the user count in the scheduled event.

        Returns
        -------
        :class:`~pincer.objects.guild.scheduled_event.ScheduledEvent`
            The scheduled event object.
        """
        data = await self._http.get(
            f"guilds/{self.id}/scheduled-events/{_id}",
            params={"with_user_count": with_user_count},
        )
        return ScheduledEvent.from_dict(data)

    async def modify_scheduled_event(
        self,
        _id: int,
        name: Optional[str] = None,
        entity_type: Optional[int] = None,
        privacy_level: Optional[int] = None,
        scheduled_start_time: Optional[datetime] = None,
        scheduled_end_time: Optional[datetime] = None,
        entity_metadata: Optional[str] = None,
        channel_id: Optional[int] = None,
        description: Optional[str] = None,
        status: Optional[int] = None,
        reason: Optional[str] = None,
    ) -> ScheduledEvent:
        """
        Modify a scheduled event.

        Parameters
        ----------
        _id : :class:`int`
            The id of the scheduled event.
        name : Optional[:class:`str`]
            The name of the scheduled event.
        entity_type : Optional[:class:`int`]
            The type of entity to be scheduled.
        privacy_level : Optional[:class:`int`]
            The privacy level of the scheduled event.
        scheduled_start_time : Optional[:class:`datetime`]
            The scheduled start time of the event.
        scheduled_end_time : Optional[:class:`datetime`]
            The scheduled end time of the event.
        entity_metadata : Optional[:class:`str`]
            The metadata of the entity to be scheduled.
        channel_id : Optional[:class:`int`]
            The channel id of the channel to be scheduled.
        description : Optional[:class:`str`]
            The description of the scheduled event.
        status : Optional[:class:`int`]
            The status of the scheduled event.
        reason : Optional[:class:`str`]
            The reason for modifying the scheduled event.

        Raises
        ------
        :class:`ValueError`
            If the scheduled event is in the past,
            or if the scheduled end time is before the scheduled start time.

        Returns
        -------
        :class:`~pincer.objects.guild.scheduled_event.ScheduledEvent`
            The scheduled event object.
        """
        if scheduled_start_time:
            if scheduled_start_time < datetime.now():
                raise ValueError("An event cannot be created in the past")

            if scheduled_end_time and scheduled_end_time < scheduled_start_time:
                raise ValueError("An event cannot start before it ends")

        kwargs: Dict[str, str] = remove_none(
            {
                "name": name,
                "privacy_level": privacy_level,
                "entity_type": entity_type,
                "scheduled_start_time": scheduled_start_time.isoformat()
                if scheduled_start_time is not None
                else None,
                "scheduled_end_time": scheduled_end_time.isoformat()
                if scheduled_end_time is not None
                else None,
                "entity_metadata": entity_metadata,
                "channel_id": channel_id,
                "description": description,
                "status": status,
            }
        )

        data = await self._http.patch(
            f"guilds/{self.id}/scheduled-events/{_id}",
            data=kwargs,
            headers={"X-Audit-Log-Reason": reason},
        )
        return ScheduledEvent.from_dict(data)

    async def delete_scheduled_event(self, _id: int):
        """
        Delete a scheduled event.

        Parameters
        ----------
        _id : :class:`int`
            The id of the scheduled event.
        """
        await self._http.delete(f"guilds/{self.id}/scheduled-events/{_id}")

    def get_guild_scheduled_event_users(
        self,
        _id: int,
        limit: int = 100,
        with_member: bool = False,
        before: Optional[int] = None,
        after: Optional[int] = None,
    ) -> APIDataGen[GuildScheduledEventUser]:
        """
        Get the users of a scheduled event.

        Parameters
        ----------
        _id : :class:`int`
            The id of the scheduled event.
        limit : :class:`int`
            The number of users to retrieve.
        with_member : :class:`bool`
            Whether to include the member object in the scheduled event user.
        before : Optional[:class:`int`]
            consider only users before given user id
        after : Optional[:class:`int`]
            consider only users after given user id

        Yields
        ------
        :class:`~pincer.objects.guild.scheduled_event.GuildScheduledEventUser`
            The scheduled event user object.
        """
        params = remove_none(
            {
                "limit": limit,
                "with_member": with_member,
                "before": before,
                "after": after,
            }
        )

        return APIDataGen(
            GuildScheduledEventUser,
            self._http.get(
                f"guilds/{self.id}/scheduled-events/{_id}/users",
                params=params,
            ),
        )

    @classmethod
    def from_dict(cls, data) -> Guild:
        """
        Parameters
        ----------
        data : :class:`Dict`
            Guild data received from the discord API.
        Returns
        -------
        :class:`~pincer.objects.guild.guild.Guild`
            The new guild object.
        Raises
        ------
        :class:`~pincer.exceptions.UnavailableGuildError`
            The guild is unavailable due to a discord outage.
        """
        if data.get("unavailable", False):
            raise UnavailableGuildError(
                f"Guild \"{data['id']}\" is unavailable due to a discord"
                " outage."
            )

        return super().from_dict(data)


@dataclass(repr=False)
class UnavailableGuild(APIObject):
    id: Snowflake
    unavailable: bool = True



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .channel import ChannelPinsUpdateEvent
from .error import DiscordError
from .gateway_commands import (
    Identify,
    Resume,
    RequestGuildMembers,
    UpdateVoiceState,
    StatusType,
    UpdatePresence,
)
from .guild import (
    GuildBanAddEvent,
    GuildBanRemoveEvent,
    GuildEmojisUpdateEvent,
    GuildStickersUpdateEvent,
    GuildIntegrationsUpdateEvent,
    GuildMemberRemoveEvent,
    GuildMemberUpdateEvent,
    GuildMembersChunkEvent,
    GuildRoleCreateEvent,
    GuildRoleUpdateEvent,
    GuildRoleDeleteEvent,
)
from .hello_ready import HelloEvent, ReadyEvent
from .integration import IntegrationDeleteEvent
from .invite import InviteCreateEvent, InviteDeleteEvent
from .message import (
    MessageDeleteEvent,
    MessageDeleteBulkEvent,
    MessageReactionAddEvent,
    MessageReactionRemoveEvent,
    MessageReactionRemoveAllEvent,
    MessageReactionRemoveEmojiEvent,
)
from .presence import (
    ActivityType,
    ActivityTimestamp,
    ActivityEmoji,
    ActivityParty,
    ActivityAssets,
    ActivitySecrets,
    ActivityFlags,
    ActivityButton,
    Activity,
    ClientStatus,
    PresenceUpdateEvent,
)
from .thread import ThreadListSyncEvent, ThreadMembersUpdateEvent
from .typing_start import TypingStartEvent
from .voice import VoiceServerUpdateEvent
from .webhook import WebhooksUpdateEvent

__all__ = (
    "Activity",
    "ActivityAssets",
    "ActivityButton",
    "ActivityEmoji",
    "ActivityFlags",
    "ActivityParty",
    "ActivitySecrets",
    "ActivityTimestamp",
    "ActivityType",
    "ChannelPinsUpdateEvent",
    "ClientStatus",
    "DiscordError",
    "GuildBanAddEvent",
    "GuildBanRemoveEvent",
    "GuildEmojisUpdateEvent",
    "GuildIntegrationsUpdateEvent",
    "GuildMemberRemoveEvent",
    "GuildMemberUpdateEvent",
    "GuildMembersChunkEvent",
    "GuildRoleCreateEvent",
    "GuildRoleDeleteEvent",
    "GuildRoleUpdateEvent",
    "GuildStickersUpdateEvent",
    "HelloEvent",
    "Identify",
    "IntegrationDeleteEvent",
    "InviteCreateEvent",
    "InviteDeleteEvent",
    "MessageDeleteBulkEvent",
    "MessageDeleteEvent",
    "MessageReactionAddEvent",
    "MessageReactionRemoveAllEvent",
    "MessageReactionRemoveEmojiEvent",
    "MessageReactionRemoveEvent",
    "PresenceUpdateEvent",
    "ReadyEvent",
    "RequestGuildMembers",
    "Resume",
    "StatusType",
    "ThreadListSyncEvent",
    "ThreadMembersUpdateEvent",
    "TypingStartEvent",
    "UpdatePresence",
    "UpdateVoiceState",
    "VoiceServerUpdateEvent",
    "WebhooksUpdateEvent",
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .dispatch import GatewayDispatch
from .gateway import Gateway, GatewayInfo
from .http import HTTPClient
from .ratelimiter import RateLimiter, Bucket


__all__ = (
    "Bucket",
    "Gateway",
    "GatewayDispatch",
    "GatewayInfo",
    "HTTPClient",
    "RateLimiter",
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import Optional, TYPE_CHECKING

from ...utils.api_object import APIObject, MISSING

if TYPE_CHECKING:
    from ..guild.stage import PrivacyLevel
    from ..guild.member import GuildMember
    from ..user.user import User
    from ...utils.snowflake import Snowflake
    from ...utils.timestamp import Timestamp
    from ...utils.types import APINullable


class EventStatus(IntEnum):
    """
    The status of an event.

    Attributes
    ----------
    SCHEDULED: int
        The event is scheduled.
    ACTIVE: int
        The event is active.
    CANCELLED: int
        The event is cancelled.
    COMPLETED: int
        The event is completed.
    """

    SCHEDULED = 1
    ACTIVE = 2
    COMPLETED = 3
    CANCELLED = 4


class GuildScheduledEventEntityType(IntEnum):
    """
    The type of entity that is scheduled.

    Attributes
    ----------
    STAGE_INSTANCE: int
        The event is scheduled for a stage instance.
    VOICE: int
        The event is scheduled for a voice channel.
    EXTERNAL: int
        The event is scheduled for an external resource.
    """

    STAGE_INSTANCE = 1
    VOICE = 2
    EXTERNAL = 3


@dataclass(repr=False)
class ScheduledEvent(APIObject):
    """
    Represents a scheduled event in a guild.

    Attributes
    ----------
    id: :class:`int`
        The ID of the scheduled event.
    name: :class:`str`
        The name of the scheduled event (1-100 characters)
    guild_id: :class:`int`
        The guild id which the scheduled event belongs to
    scheduled_start_time: str
        The scheduled start time of the event.
    privacy_level: :class:`~pincer.guild.stage.PrivacyLevel`
        The privacy level of the scheduled event.
    status: :class:`~pincer.guild.schedule_events.EventStatus`
        The status of the scheduled event.
    entity_type: :class:`~pincer.guild.schedule_events.GuildScheduledEventEntityType`
        The type of the scheduled event
    channel_id: APINullable[:class:`int`]
        The channel id in which the scheduled event will be hosted,
        or null if scheduled entity type is EXTERNAL
    creator_id: APINullable[:class:`int`]
        The user id of the creator of the scheduled event
    scheduled_end_time: str
        The time the scheduled event will end, required if entity_type is EXTERNAL
    description: APINullable[:class:`str`]
        The description of the scheduled event (0-1000 characters)
    entity_id: APINullable[:class:`int`]
        The id of an entity associated with a guild scheduled event
    entity_metadata: APINullable[:class:`str`]
        Additional metadata for the guild scheduled event
    creator: APINullable[:class:`~pincer.objects.user.user.User`]
        The user who created the scheduled event
    user_count: APINullable[:class:`int`]
        The number of users who have joined the scheduled event
    """

    id: Snowflake
    name: str
    guild_id: Snowflake
    scheduled_start_time: Timestamp
    privacy_level: PrivacyLevel
    status: EventStatus
    entity_type: GuildScheduledEventEntityType

    channel_id: APINullable[Snowflake] = MISSING
    creator_id: APINullable[Snowflake] = MISSING
    scheduled_end_time: Optional[Timestamp] = None

    description: APINullable[str] = MISSING
    entity_id: APINullable[Snowflake] = MISSING
    entity_metadata: APINullable[str] = MISSING
    creator: APINullable[User] = MISSING
    user_count: APINullable[int] = MISSING


@dataclass
class GuildScheduledEventUser(APIObject):
    """
    Represents a user who has joined a scheduled event.

    Attributes
    ----------
    guild_scheduled_event_id: :class:`int`
        the scheduled event id which the user subscribed to
    user : :class:`~pincer.objects.user.user.User`
        user which subscribed to an event
    member : APINullable[:class:`~pincer.objects.guild.member.GuildMember`]
                guild member data for this user for the guild which this event belongs to, if any
    """

    guild_scheduled_event_id: Snowflake
    user: User
    member: APINullable[GuildMember] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from typing import Any, Callable, Iterable, List, T


def replace(
    func: Callable[[Any], bool], iter_: Iterable[T], new_item: T
) -> List[T]:
    return [item if func(item) else new_item for item in iter_]
